from __future__ import annotations

import time

from functools import wraps
from typing import TYPE_CHECKING, Any, ClassVar

from memos.log import get_logger
from memos.mem_scheduler.general_scheduler import GeneralScheduler
from memos.mem_scheduler.schemas.general_schemas import (
    DEFAULT_MAX_QUERY_KEY_WORDS,
    UserID,
)
from memos.mem_scheduler.schemas.monitor_schemas import QueryMonitorItem


if TYPE_CHECKING:
    from memos.memories.textual.tree import TextualMemoryItem


logger = get_logger(__name__)


class SchedulerForEval(GeneralScheduler):
    """
    A scheduler class that inherits from GeneralScheduler and provides evaluation-specific functionality.
    This class extends GeneralScheduler with evaluation methods.
    """

    # Class variable to store timing information for all instances
    timer_cache: ClassVar[dict[str, dict[str, Any]]] = {}

    def __init__(self, config):
        """
        Initialize the SchedulerForEval with the same configuration as GeneralScheduler.

        Args:
            config: Configuration object for the scheduler
        """
        super().__init__(config)
        # Initialize instance timer_cache
        self.timer_cache = {}

    @staticmethod
    def time_it(func_name: str | None = None):
        """
        Static method decorator to measure function execution time and store in timer_cache.

        Args:
            func_name: Custom name for the function in timer_cache. If None, uses function.__name__
        """

        def decorator(func):
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                # Get function name
                name = func_name or func.__name__

                # Start timing
                start_time = time.time()
                result = func(self, *args, **kwargs)
                end_time = time.time()

                # Calculate execution time
                exec_time = end_time - start_time

                # Format time as HH:MM:SS.mmm
                hours = int(exec_time // 3600)
                minutes = int((exec_time % 3600) // 60)
                seconds = exec_time % 60

                if hours > 0:
                    time_str = f"{hours:02d}:{minutes:02d}:{seconds:06.3f}"
                else:
                    time_str = f"{minutes:02d}:{seconds:06.3f}"

                # Store in timer_cache
                if not hasattr(self, "timer_cache"):
                    self.timer_cache = {}

                self.timer_cache[name] = {
                    "time_str": time_str,
                    "seconds": exec_time,
                }

                logger.info(f"{name} executed in {time_str}")
                return result

            return wrapper

        return decorator

    def get_timer_summary(self) -> str:
        """
        Get a summary of all timed functions.

        Returns:
            Formatted string with timing information
        """
        if not self.timer_cache:
            return "No timing data available."

        summary = "=== Timing Summary ===\n"
        for func_name, data in self.timer_cache.items():
            summary += f"{func_name}: {data['time_str']} (at {data['timestamp']})\n"

        return summary

    def clear_timer_cache(self):
        """Clear the timer cache."""
        self.timer_cache.clear()

    @time_it("update_working_memory")
    def update_working_memory_for_eval(
        self, query: str, user_id: UserID | str, top_k: int
    ) -> list[str]:
        """
        Update working memory based on query and return the updated memory list.

        Args:
            query: The query string
            user_id: User identifier
            top_k: Number of top memories to return

        Returns:
            List of memory strings from updated working memory
        """
        self.monitor.register_query_monitor_if_not_exists(
            user_id=user_id, mem_cube_id=self.current_mem_cube_id
        )

        query_keywords = self.monitor.extract_query_keywords(query=query)
        logger.info(f'Extract keywords "{query_keywords}" from query "{query}"')

        item = QueryMonitorItem(
            user_id=user_id,
            mem_cube_id=self.current_mem_cube_id,
            query_text=query,
            keywords=query_keywords,
            max_keywords=DEFAULT_MAX_QUERY_KEY_WORDS,
        )
        query_db_manager = self.monitor.query_monitors[user_id][self.current_mem_cube_id]
        query_db_manager.obj.put(item=item)
        # Sync with database after adding new item
        query_db_manager.sync_with_orm()
        logger.debug(f"Queries in monitor are {query_db_manager.obj.get_queries_with_timesort()}.")

        queries = [query]

        # recall
        mem_cube = self.current_mem_cube
        text_mem_base = mem_cube.text_mem

        cur_working_memory: list[TextualMemoryItem] = text_mem_base.get_working_memory()
        text_working_memory: list[str] = [w_m.memory for w_m in cur_working_memory]
        intent_result = self.monitor.detect_intent(
            q_list=queries, text_working_memory=text_working_memory
        )

        if intent_result["trigger_retrieval"]:
            missing_evidences = intent_result["missing_evidences"]
            num_evidence = len(missing_evidences)
            k_per_evidence = max(1, top_k // max(1, num_evidence))
            new_candidates = []
            for item in missing_evidences:
                logger.info(f"missing_evidences: {item}")
                results: list[TextualMemoryItem] = self.retriever.search(
                    query=item,
                    mem_cube=mem_cube,
                    top_k=k_per_evidence,
                    method=self.search_method,
                )
                logger.info(
                    f"search results for {missing_evidences}: {[one.memory for one in results]}"
                )
                new_candidates.extend(results)
            logger.info(
                f"missing_evidences: {missing_evidences} and get {len(new_candidates)} new candidate memories."
            )
        else:
            new_candidates = []
            logger.info(f"intent_result: {intent_result}. not triggered")

        # rerank
        new_order_working_memory = self.replace_working_memory(
            user_id=user_id,
            mem_cube_id=self.current_mem_cube_id,
            mem_cube=self.current_mem_cube,
            original_memory=cur_working_memory,
            new_memory=new_candidates,
        )
        new_order_working_memory = new_order_working_memory[:top_k]
        logger.info(f"size of new_order_working_memory: {len(new_order_working_memory)}")

        return [m.memory for m in new_order_working_memory]

    @time_it("memory_answer_ability")
    def evaluate_memory_answer_ability(
        self, query: str, memory_texts: list[str], top_k: int = 100
    ) -> bool:
        """
        Use LLM to evaluate whether the given memories can answer the query.

        Args:
            query: The query string to evaluate
            memory_texts: List of memory texts to check against
            top_k: Maximum number of memories to consider for evaluation

        Returns:
            Boolean indicating whether the memories can answer the query
        """
        # Limit the number of memories to evaluate
        limited_memories = memory_texts[:top_k] if memory_texts else []

        # Build prompt using the template
        prompt = self.monitor.build_prompt(
            template_name="memory_answer_ability_evaluation",
            query=query,
            memory_list="\n".join([f"- {memory}" for memory in limited_memories])
            if limited_memories
            else "No memories available",
        )

        # Use the process LLM to generate response
        response = self.monitor._process_llm.generate([{"role": "user", "content": prompt}])

        try:
            # Extract JSON response
            from memos.mem_scheduler.utils.misc_utils import extract_json_obj

            result = extract_json_obj(response)

            # Validate response structure
            if "result" in result:
                logger.info(
                    f"Memory answer ability evaluation result: {result['result']}, reason: {result.get('reason', 'No reason provided')}"
                )
                return result["result"]
            else:
                logger.warning(f"Invalid response structure from LLM: {result}")
                return False

        except Exception as e:
            logger.error(
                f"Failed to parse LLM response for memory answer ability evaluation: {response}. Error: {e}"
            )
            # Fallback: return False if we can't determine answer ability
            return False

    @time_it("search_for_eval")
    def search_for_eval(
        self, query: str, user_id: UserID | str, top_k: int, scheduler_flag: bool = True
    ) -> list[str]:
        """
        Original search_for_eval function refactored to use the new decomposed functions.

        Args:
            query: The query string
            user_id: User identifier
            top_k: Number of top memories to return
            scheduler_flag: Whether to update working memory or just evaluate

        Returns:
            Tuple of (memory_list, can_answer_boolean)
        """
        if not scheduler_flag:
            # Get current working memory without updating
            mem_cube = self.current_mem_cube
            text_mem_base = mem_cube.text_mem
            cur_working_memory: list[TextualMemoryItem] = text_mem_base.get_working_memory()
            text_working_memory: list[str] = [w_m.memory for w_m in cur_working_memory]

            return text_working_memory
        else:
            # Update working memory and get the result
            updated_memories = self.update_working_memory_for_eval(
                query=query, user_id=user_id, top_k=top_k
            )

            return updated_memories



import time

from memos.log import get_logger


logger = get_logger(__name__)


def timed(func=None, *, log=True, log_prefix=""):
    """Decorator to measure and optionally log time of retrieval steps.

    Can be used as @timed or @timed(log=True)
    """

    def decorator(fn):
        def wrapper(*args, **kwargs):
            start = time.perf_counter()
            result = fn(*args, **kwargs)
            elapsed = time.perf_counter() - start
            elapsed_ms = elapsed * 1000.0
            if log:
                logger.info(f"[TIMER] {log_prefix or fn.__name__} took {elapsed_ms:.0f} ms")
            return result

        return wrapper

    # Handle both @timed and @timed(log=True) cases
    if func is None:
        return decorator
    return decorator(func)



import os

from enum import Enum
from pathlib import Path
from typing import NewType


class SearchMode(str, Enum):
    """Enumeration for search modes."""

    FAST = "fast"
    FINE = "fine"
    MIXTURE = "mixture"


class FineStrategy(str, Enum):
    """Enumeration for fine strategies."""

    REWRITE = "rewrite"
    RECREATE = "recreate"
    DEEP_SEARCH = "deep_search"


FILE_PATH = Path(__file__).absolute()
BASE_DIR = FILE_PATH.parent.parent.parent.parent.parent

QUERY_LABEL = "query"
ANSWER_LABEL = "answer"
ADD_LABEL = "add"
MEM_READ_LABEL = "mem_read"
MEM_ORGANIZE_LABEL = "mem_organize"
API_MIX_SEARCH_LABEL = "api_mix_search"
PREF_ADD_LABEL = "pref_add"

TreeTextMemory_SEARCH_METHOD = "tree_text_memory_search"
TreeTextMemory_FINE_SEARCH_METHOD = "tree_text_memory_fine_search"
TextMemory_SEARCH_METHOD = "text_memory_search"
DIRECT_EXCHANGE_TYPE = "direct"
FANOUT_EXCHANGE_TYPE = "fanout"
DEFAULT_WORKING_MEM_MONITOR_SIZE_LIMIT = 30
DEFAULT_ACTIVATION_MEM_MONITOR_SIZE_LIMIT = 20
DEFAULT_ACT_MEM_DUMP_PATH = f"{BASE_DIR}/outputs/mem_scheduler/mem_cube_scheduler_test.kv_cache"
DEFAULT_THREAD_POOL_MAX_WORKERS = 50
DEFAULT_CONSUME_INTERVAL_SECONDS = 0.01
DEFAULT_CONSUME_BATCH = 1
DEFAULT_DISPATCHER_MONITOR_CHECK_INTERVAL = 300
DEFAULT_DISPATCHER_MONITOR_MAX_FAILURES = 2
DEFAULT_STUCK_THREAD_TOLERANCE = 10
DEFAULT_MAX_INTERNAL_MESSAGE_QUEUE_SIZE = -1
DEFAULT_TOP_K = 10
DEFAULT_CONTEXT_WINDOW_SIZE = 5
DEFAULT_USE_REDIS_QUEUE = True
DEFAULT_MULTI_TASK_RUNNING_TIMEOUT = 30
DEFAULT_SCHEDULER_RETRIEVER_BATCH_SIZE = 20
DEFAULT_SCHEDULER_RETRIEVER_RETRIES = 1
DEFAULT_STOP_WAIT = False

# startup mode configuration
STARTUP_BY_THREAD = "thread"
STARTUP_BY_PROCESS = "process"
DEFAULT_STARTUP_MODE = STARTUP_BY_THREAD  # default to thread mode

NOT_INITIALIZED = -1


# web log
LONG_TERM_MEMORY_TYPE = "LongTermMemory"
USER_MEMORY_TYPE = "UserMemory"
WORKING_MEMORY_TYPE = "WorkingMemory"
TEXT_MEMORY_TYPE = "TextMemory"
ACTIVATION_MEMORY_TYPE = "ActivationMemory"
PARAMETER_MEMORY_TYPE = "ParameterMemory"
USER_INPUT_TYPE = "UserInput"
NOT_APPLICABLE_TYPE = "NotApplicable"

# monitors
MONITOR_WORKING_MEMORY_TYPE = "MonitorWorkingMemoryType"
MONITOR_ACTIVATION_MEMORY_TYPE = "MonitorActivationMemoryType"
DEFAULT_MAX_QUERY_KEY_WORDS = 1000
DEFAULT_WEIGHT_VECTOR_FOR_RANKING = [0.9, 0.05, 0.05]
DEFAULT_MAX_WEB_LOG_QUEUE_SIZE = 50


# new types
UserID = NewType("UserID", str)
MemCubeID = NewType("CubeID", str)

# algorithm strategies
DEFAULT_FINE_STRATEGY = FineStrategy.REWRITE

# Read fine strategy from environment variable `FINE_STRATEGY`.
# If provided and valid, use it; otherwise fall back to default.
_env_fine_strategy = os.getenv("FINE_STRATEGY")
if _env_fine_strategy:
    try:
        FINE_STRATEGY = FineStrategy(_env_fine_strategy)
    except ValueError:
        FINE_STRATEGY = DEFAULT_FINE_STRATEGY
else:
    FINE_STRATEGY = DEFAULT_FINE_STRATEGY



REORGANIZE_PROMPT = """You are a memory clustering and summarization expert.

Given the following child memory items:

{memory_items_text}

Please perform:
1. Identify information that reflects user's experiences, beliefs, concerns, decisions, plans, or reactions — including meaningful input from assistant that user acknowledged or responded to.
2. Resolve all time, person, and event references clearly:
   - Convert relative time expressions (e.g., “yesterday,” “next Friday”) into absolute dates using the message timestamp if possible.
   - Clearly distinguish between event time and message time.
   - If uncertainty exists, state it explicitly (e.g., “around June 2025,” “exact date unclear”).
   - Include specific locations if mentioned.
   - Resolve all pronouns, aliases, and ambiguous references into full names or identities.
   - Disambiguate people with the same name if applicable.
3. Always write from a third-person perspective, referring to user as
"The user" or by name if name mentioned, rather than using first-person ("I", "me", "my").
For example, write "The user felt exhausted..." instead of "I felt exhausted...".
4. Do not omit any information that user is likely to remember.
   - Include all key experiences, thoughts, emotional responses, and plans — even if they seem minor.
   - Prioritize completeness and fidelity over conciseness.
   - Do not generalize or skip details that could be personally meaningful to user.
5. Summarize all child memory items into one memory item.

Language rules:
- The `key`, `value`, `tags`, `summary` fields must match the mostly used language of the input memory items.  **如果输入是中文，请输出中文**
- Keep `memory_type` in English.

Return valid JSON:
{
  "key": <string, a concise title of the `value` field>,
  "memory_type": <string, Either "LongTermMemory" or "UserMemory">,
  "value": <A detailed, self-contained, and unambiguous memory statement, only contain detailed, unaltered information extracted and consolidated from the input `value` fields, do not include summary content — written in English if the input memory items are in English, or in Chinese if the input is in Chinese>,
  "tags": <A list of relevant thematic keywords (e.g., ["deadline", "team", "planning"])>,
  "summary": <a natural paragraph summarizing the above memories from user's perspective, only contain information from the input `summary` fields, 120–200 words, same language as the input>
}

"""

DOC_REORGANIZE_PROMPT = """You are a document summarization and knowledge extraction expert.

Given the following summarized document items:

{memory_items_text}

Please perform:
1. Identify key information that reflects factual content, insights, decisions, or implications from the documents — including any notable themes, conclusions, or data points.
2. Resolve all time, person, location, and event references clearly:
   - Convert relative time expressions (e.g., “last year,” “next quarter”) into absolute dates if context allows.
   - Clearly distinguish between event time and document time.
   - If uncertainty exists, state it explicitly (e.g., “around 2024,” “exact date unclear”).
   - Include specific locations if mentioned.
   - Resolve all pronouns, aliases, and ambiguous references into full names or identities.
   - Disambiguate entities with the same name if applicable.
3. Always write from a third-person perspective, referring to the subject or content clearly rather than using first-person ("I", "me", "my").
4. Do not omit any information that is likely to be important or memorable from the document summaries.
   - Include all key facts, insights, emotional tones, and plans — even if they seem minor.
   - Prioritize completeness and fidelity over conciseness.
   - Do not generalize or skip details that could be contextually meaningful.
5. Summarize all document summaries into one integrated memory item.

Language rules:
- The `key`, `value`, `tags`, `summary` fields must match the mostly used language of the input document summaries.  **如果输入是中文，请输出中文**
- Keep `memory_type` in English.

Return valid JSON:
{
  "key": <string, a concise title of the `value` field>,
  "memory_type": "LongTermMemory",
  "value": <A detailed, self-contained, and unambiguous memory statement, only contain detailed, unaltered information extracted and consolidated from the input `value` fields, do not include summary content — written in English if the input memory items are in English, or in Chinese if the input is in Chinese>,
  "tags": <A list of relevant thematic keywords (e.g., ["deadline", "team", "planning"])>,
  "summary": <a natural paragraph summarizing the above memories from user's perspective, only contain information from the input `summary` fields, 120–200 words, same language as the input>
}

"""


LOCAL_SUBCLUSTER_PROMPT = """You are a memory organization expert.

You are given a cluster of memory items, each with an ID and content.
Your task is to divide these into smaller, semantically meaningful sub-clusters.

Instructions:
- Identify natural topics by analyzing common time, place, people, and event elements.
- Each sub-cluster must reflect a coherent theme that helps retrieval.
- Each sub-cluster should have 2–10 items. Discard singletons.
- Each item ID must appear in exactly one sub-cluster or be discarded. No duplicates are allowed.
- All IDs in the output must be from the provided Memory items.
- Return strictly valid JSON only.

Example: If you have items about a project across multiple phases, group them by milestone, team, or event.

Language rules:
- The `key` fields must match the mostly used language of the clustered memories. **如果输入是中文，请输出中文**

Return valid JSON:
{
  "clusters": [
    {
      "ids": ["<id1>", "<id2>", ...],
      "key": "<string, a unique, concise memory title>"
    },
    ...
  ]
}

Memory items:
{joined_scene}
"""

PAIRWISE_RELATION_PROMPT = """
You are a reasoning assistant.

Given two memory units:
- Node 1: "{node1}"
- Node 2: "{node2}"

Your task:
- Determine their relationship ONLY if it reveals NEW usable reasoning or retrieval knowledge that is NOT already explicit in either unit.
- Focus on whether combining them adds new temporal, causal, conditional, or conflict information.

Valid options:
- CAUSE: One clearly leads to the other.
- CONDITION: One happens only if the other condition holds.
- RELATE: They are semantically related by shared people, time, place, or event, but neither causes the other.
- CONFLICT: They logically contradict each other.
- NONE: No clear useful connection.

Example:
- Node 1: "The marketing campaign ended in June."
- Node 2: "Product sales dropped in July."
Answer: CAUSE

Another Example:
- Node 1: "The conference was postponed to August due to the venue being unavailable."
- Node 2: "The venue was booked for a wedding in August."
Answer: CONFLICT

Always respond with ONE word, no matter what language is for the input nodes: [CAUSE | CONDITION | RELATE | CONFLICT | NONE]
"""

INFER_FACT_PROMPT = """
You are an inference expert.

Source Memory: "{source}"
Target Memory: "{target}"

They are connected by a {relation_type} relation.
Derive ONE new factual statement that clearly combines them in a way that is NOT a trivial restatement.

Requirements:
- Include relevant time, place, people, and event details if available.
- If the inference is a logical guess, explicitly use phrases like "It can be inferred that...".

Example:
Source: "John missed the team meeting on Monday."
Target: "Important project deadlines were discussed in that meeting."
Relation: CAUSE
Inference: "It can be inferred that John may not know the new project deadlines."

If there is NO new useful fact that combines them, reply exactly: "None"
"""

AGGREGATE_PROMPT = """
You are a concept summarization assistant.

Below is a list of memory items:
{joined}

Your task:
- Identify if they can be meaningfully grouped under a new, higher-level concept that clarifies their shared time, place, people, or event context.
- Do NOT aggregate if the overlap is trivial or obvious from each unit alone.
- If the summary involves any plausible interpretation, explicitly note it (e.g., "This suggests...").

Example:
Input Memories:
- "Mary organized the 2023 sustainability summit in Berlin."
- "Mary presented a keynote on renewable energy at the same summit."

Language rules:
- The `key`, `value`, `tags`, `background` fields must match the language of the input.

Good Aggregate:
{
  "key": "Mary's Sustainability Summit Role",
  "value": "Mary organized and spoke at the 2023 sustainability summit in Berlin, highlighting renewable energy initiatives.",
  "tags": ["Mary", "summit", "Berlin", "2023"],
  "background": "Combined from multiple memories about Mary's activities at the summit."
}

If you find NO useful higher-level concept, reply exactly: "None".
"""

REDUNDANCY_MERGE_PROMPT = """You are given two pieces of text joined by the marker `⟵MERGED⟶`. Please carefully read both sides of the merged text. Your task is to summarize and consolidate all the factual details from both sides into a single, coherent text, without omitting any information. You must include every distinct detail mentioned in either text. Do not provide any explanation or analysis — only return the merged summary. Don't use pronouns or subjective language, just the facts as they are presented.\n{merged_text}"""


MEMORY_RELATION_DETECTOR_PROMPT = """You are a memory relationship analyzer.
You are given two plaintext statements. Determine the relationship between them. Classify the relationship into one of the following categories:

contradictory: The two statements describe the same event or related aspects of it but contain factually conflicting details.
redundant: The two statements describe essentially the same event or information with significant overlap in content and details, conveying the same core information (even if worded differently).
independent: The two statements are either about different events/topics (unrelated) OR describe different, non-overlapping aspects or perspectives of the same event without conflict (complementary). In both sub-cases, they provide distinct information without contradiction.
Respond only with one of the three labels: contradictory, redundant, or independent.
Do not provide any explanation or additional text.

Statement 1: {statement_1}
Statement 2: {statement_2}
"""


MEMORY_RELATION_RESOLVER_PROMPT = """You are a memory fusion expert. You are given two statements and their associated metadata. The statements have been identified as {relation}. Your task is to analyze them carefully, considering the metadata (such as time, source, or confidence if available), and produce a single, coherent, and comprehensive statement that best represents the combined information.

If the statements are redundant, merge them by preserving all unique details and removing duplication, forming a richer, consolidated version.
If the statements are contradictory, attempt to resolve the conflict by prioritizing more recent information, higher-confidence data, or logically reconciling the differences based on context. If the contradiction is fundamental and cannot be logically resolved, output <answer>No</answer>.
Do not include any explanations, reasoning, or extra text. Only output the final result enclosed in <answer></answer> tags.
Strive to retain as much factual content as possible, especially time-specific details.
Use objective language and avoid pronouns.
Output Example 1 (unresolvable conflict):
<answer>No</answer>

Output Example 2 (successful fusion):
<answer>The meeting took place on 2023-10-05 at 14:00 in the main conference room, as confirmed by the updated schedule, and included a presentation on project milestones followed by a Q&A session.</answer>

Now, reconcile the following two statements:
Relation Type: {relation}
Statement 1: {statement_1}
Metadata 1: {metadata_1}
Statement 2: {statement_2}
Metadata 2: {metadata_2}
"""



"""User management system for MemOS.

This module provides user authentication, authorization, and cube management
functionality using SQLAlchemy and SQLite.
"""

import uuid

from datetime import datetime
from enum import Enum
from pathlib import Path

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    String,
    Table,
    create_engine,
)
from sqlalchemy import (
    Enum as SQLEnum,
)
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, declarative_base, relationship, sessionmaker

from memos import settings
from memos.log import get_logger


logger = get_logger(__name__)

Base = declarative_base()


class UserRole(Enum):
    """User roles enumeration."""

    ROOT = "ROOT"
    ADMIN = "ADMIN"
    USER = "USER"
    GUEST = "GUEST"


# Association table for many-to-many relationship between users and cubes
user_cube_association = Table(
    "user_cube_association",
    Base.metadata,
    Column("user_id", String, ForeignKey("users.user_id"), primary_key=True),
    Column("cube_id", String, ForeignKey("cubes.cube_id"), primary_key=True),
    Column("created_at", DateTime, default=datetime.now),
)


class User(Base):
    """User model for the database."""

    __tablename__ = "users"

    user_id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    user_name = Column(String, unique=True, nullable=False)
    role = Column(SQLEnum(UserRole), default=UserRole.USER, nullable=False)
    created_at = Column(DateTime, default=datetime.now, nullable=False)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now, nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)

    # Relationship with cubes
    cubes = relationship("Cube", secondary=user_cube_association, back_populates="users")
    owned_cubes = relationship("Cube", back_populates="owner", cascade="all, delete-orphan")

    def __repr__(self):
        return f"<User(user_id='{self.user_id}', user_name='{self.user_name}', role='{self.role.value}')>"


class Cube(Base):
    """Cube model for the database."""

    __tablename__ = "cubes"

    cube_id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    cube_name = Column(String, nullable=False)
    cube_path = Column(String, nullable=True)  # Local path or remote repo
    owner_id = Column(String, ForeignKey("users.user_id"), nullable=False)
    created_at = Column(DateTime, default=datetime.now, nullable=False)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now, nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)

    # Relationships
    owner = relationship("User", back_populates="owned_cubes")
    users = relationship("User", secondary=user_cube_association, back_populates="cubes")

    def __repr__(self):
        return f"<Cube(cube_id='{self.cube_id}', cube_name='{self.cube_name}', owner_id='{self.owner_id}')>"


class UserManager:
    """User management system for MemOS."""

    def __init__(self, db_path: str | None = None, user_id: str = "root"):
        """Initialize the user manager with database connection.

        Args:
            db_path (str, optional): Path to the SQLite database file.
                If None, uses default path in MEMOS_DIR.
            user_id (str, optional): User ID. If None, uses default user ID.
        """
        if db_path is None:
            db_path = str(settings.MEMOS_DIR / "memos_users.db")

        # Ensure the directory exists
        Path(db_path).parent.mkdir(parents=True, exist_ok=True)

        self.db_path = db_path
        self.engine = create_engine(f"sqlite:///{db_path}", echo=False)
        self.SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)

        # Create tables
        Base.metadata.create_all(bind=self.engine)

        # Initialize with root user if no users exist
        self._init_root_user(user_id)

        logger.info(f"UserManager initialized with database at {db_path}")

    def _get_session(self) -> Session:
        """Get a database session."""
        return self.SessionLocal()

    def _init_root_user(self, user_id: str) -> None:
        """Initialize the root user if no users exist."""
        session = self._get_session()
        try:
            # Check if any users exist
            user_count = session.query(User).count()
            if user_count == 0:
                root_user = User(user_id=user_id, user_name=user_id, role=UserRole.ROOT)
                session.add(root_user)
                session.commit()
                logger.info("Root user created successfully")
            else:
                self.create_user(user_name=user_id, user_id=user_id, role=UserRole.ROOT)
        except Exception as e:
            session.rollback()
            logger.error(f"Failed to create {user_id} user: {e}")
        finally:
            session.close()

    def create_user(
        self, user_name: str, role: UserRole = UserRole.USER, user_id: str | None = None
    ) -> str:
        """Create a new user.

        Args:
            user_name (str): Name of the user.
            role (UserRole): Role of the user.
            user_id (str, optional): Custom user ID. If None, generates UUID.

        Returns:
            str: The created user ID.

        Raises:
            ValueError: If user_name already exists.
        """
        session = self._get_session()
        try:
            # Check if user_name already exists
            existing_user = session.query(User).filter(User.user_name == user_name).first()
            if existing_user:
                logger.info(f"User with name '{user_name}' already exists")
                return existing_user.user_id
            user = User(user_name=user_name, role=role, user_id=user_id or str(uuid.uuid4()))
            session.add(user)
            session.commit()
            logger.info(f"User '{user_name}' created with ID: {user.user_id}")
            return user.user_id
        except IntegrityError:
            session.rollback()
            logger.info(f"failed to create user with name '{user_name}' already exists")
        except Exception as e:
            session.rollback()
            logger.error(f"Error creating user: {e}")
            raise
        finally:
            session.close()

    def get_user(self, user_id: str) -> User | None:
        """Get user by ID.

        Args:
            user_id (str): The user ID.

        Returns:
            User: The user object or None if not found.
        """
        session = self._get_session()
        try:
            return session.query(User).filter(User.user_id == user_id).first()
        finally:
            session.close()

    def get_user_by_name(self, user_name: str) -> User | None:
        """Get user by name.

        Args:
            user_name (str): The user name.

        Returns:
            User: The user object or None if not found.
        """
        session = self._get_session()
        try:
            return session.query(User).filter(User.user_name == user_name).first()
        finally:
            session.close()

    def validate_user(self, user_id: str) -> bool:
        """Validate if a user exists and is active.

        Args:
            user_id (str): The user ID to validate.

        Returns:
            bool: True if user exists and is active, False otherwise.
        """
        user = self.get_user(user_id)
        return user is not None and user.is_active

    def list_users(self) -> list[User]:
        """List all active users.

        Returns:
            list[User]: List of all active users.
        """
        session = self._get_session()
        try:
            return session.query(User).filter(User.is_active).all()
        finally:
            session.close()

    def create_cube(
        self,
        cube_name: str,
        owner_id: str,
        cube_path: str | None = None,
        cube_id: str | None = None,
    ) -> str:
        """Create a new cube.

        Args:
            cube_name (str): Name of the cube.
            owner_id (str): ID of the cube owner.
            cube_path (str, optional): Path to the cube.
            cube_id (str, optional): Custom cube ID. If None, generates UUID.

        Returns:
            str: The created cube ID.

        Raises:
            ValueError: If owner doesn't exist.
        """
        session = self._get_session()
        try:
            # Validate owner exists
            owner = session.query(User).filter(User.user_id == owner_id).first()
            if not owner:
                raise ValueError(f"User with ID '{owner_id}' does not exist")

            cube = Cube(
                cube_name=cube_name,
                owner_id=owner_id,
                cube_path=cube_path,
                cube_id=cube_id or str(uuid.uuid4()),
            )
            session.add(cube)

            # Add owner to cube users
            cube.users.append(owner)

            session.commit()
            logger.info(f"Cube '{cube_name}' created with ID: {cube.cube_id}")
            return cube.cube_id
        except Exception as e:
            session.rollback()
            logger.error(f"Error creating cube: {e}")
            raise
        finally:
            session.close()

    def get_cube(self, cube_id: str) -> Cube | None:
        """Get cube by ID.

        Args:
            cube_id (str): The cube ID.

        Returns:
            Cube: The cube object or None if not found.
        """
        session = self._get_session()
        try:
            return session.query(Cube).filter(Cube.cube_id == cube_id).first()
        finally:
            session.close()

    def validate_user_cube_access(self, user_id: str, cube_id: str) -> bool:
        """Validate if a user has access to a cube.

        Args:
            user_id (str): The user ID.
            cube_id (str): The cube ID.

        Returns:
            bool: True if user has access to cube, False otherwise.
        """
        session = self._get_session()
        try:
            # Check if user exists and is active
            user = session.query(User).filter(User.user_id == user_id, User.is_active).first()
            if not user:
                return False

            # Check if cube exists and is active
            cube = session.query(Cube).filter(Cube.cube_id == cube_id, Cube.is_active).first()
            if not cube:
                return False

            # Check if user has access to cube (owner or in users list)
            if cube.owner_id == user_id:
                return True

            # Check many-to-many relationship
            return user in cube.users
        finally:
            session.close()

    def get_user_cubes(self, user_id: str) -> list[Cube]:
        """Get all cubes accessible by a user.

        Args:
            user_id (str): The user ID.

        Returns:
            list[Cube]: List of cubes accessible by the user.
        """
        session = self._get_session()
        try:
            user = session.query(User).filter(User.user_id == user_id).first()
            if not user:
                return []

            active_cubes = [cube for cube in user.cubes if cube.is_active]
            return sorted(active_cubes, key=lambda cube: cube.created_at, reverse=True)
        finally:
            session.close()

    def add_user_to_cube(self, user_id: str, cube_id: str) -> bool:
        """Add a user to a cube's access list.

        Args:
            user_id (str): The user ID.
            cube_id (str): The cube ID.

        Returns:
            bool: True if successful, False otherwise.
        """
        session = self._get_session()
        try:
            user = session.query(User).filter(User.user_id == user_id).first()
            cube = session.query(Cube).filter(Cube.cube_id == cube_id).first()

            if not user or not cube:
                return False

            if user not in cube.users:
                cube.users.append(user)
                session.commit()
                logger.info(f"User '{user_id}' added to cube '{cube_id}'")

            return True
        except Exception as e:
            session.rollback()
            logger.error(f"Error adding user to cube: {e}")
            return False
        finally:
            session.close()

    def remove_user_from_cube(self, user_id: str, cube_id: str) -> bool:
        """Remove a user from a cube's access list.

        Args:
            user_id (str): The user ID.
            cube_id (str): The cube ID.

        Returns:
            bool: True if successful, False otherwise.
        """
        session = self._get_session()
        try:
            user = session.query(User).filter(User.user_id == user_id).first()
            cube = session.query(Cube).filter(Cube.cube_id == cube_id).first()

            if not user or not cube:
                return False

            # Don't remove owner
            if cube.owner_id == user_id:
                logger.warning(f"Cannot remove owner '{user_id}' from cube '{cube_id}'")
                return False

            if user in cube.users:
                cube.users.remove(user)
                session.commit()
                logger.info(f"User '{user_id}' removed from cube '{cube_id}'")

            return True
        except Exception as e:
            session.rollback()
            logger.error(f"Error removing user from cube: {e}")
            return False
        finally:
            session.close()

    def delete_user(self, user_id: str) -> bool:
        """Soft delete a user (set is_active to False).

        Args:
            user_id (str): The user ID.

        Returns:
            bool: True if successful, False otherwise.
        """
        session = self._get_session()
        try:
            user = session.query(User).filter(User.user_id == user_id).first()
            if not user:
                return False

            # Don't delete root user
            if user.role == UserRole.ROOT:
                logger.warning("Cannot delete root user")
                return False

            user.is_active = False
            session.commit()
            logger.info(f"User '{user_id}' deactivated")
            return True
        except Exception as e:
            session.rollback()
            logger.error(f"Error deleting user: {e}")
            return False
        finally:
            session.close()

    def delete_cube(self, cube_id: str) -> bool:
        """Soft delete a cube (set is_active to False).

        Args:
            cube_id (str): The cube ID.

        Returns:
            bool: True if successful, False otherwise.
        """
        session = self._get_session()
        try:
            cube = session.query(Cube).filter(Cube.cube_id == cube_id).first()
            if not cube:
                return False

            cube.is_active = False
            session.commit()
            logger.info(f"Cube '{cube_id}' deactivated")
            return True
        except Exception as e:
            session.rollback()
            logger.error(f"Error deleting cube: {e}")
            return False
        finally:
            session.close()

    def close(self) -> None:
        """Close the database engine and dispose of all connections.

        This method should be called when the UserManager is no longer needed
        to ensure proper cleanup of database connections.
        """
        if hasattr(self, "engine"):
            self.engine.dispose()
            logger.info("UserManager database connections closed")



from memos.memories.textual.item import (
    TextualMemoryItem,
)


def split_continuous_references(text: str) -> str:
    """
    Split continuous reference tags into individual reference tags.

    Converts patterns like [1:92ff35fb, 4:bfe6f044] to [1:92ff35fb] [4:bfe6f044]

    Only processes text if:
    1. '[' appears exactly once
    2. ']' appears exactly once
    3. Contains commas between '[' and ']'

    Args:
        text (str): Text containing reference tags

    Returns:
        str: Text with split reference tags, or original text if conditions not met
    """
    # Early return if text is empty
    if not text:
        return text
    # Check if '[' appears exactly once
    if text.count("[") != 1:
        return text
    # Check if ']' appears exactly once
    if text.count("]") != 1:
        return text
    # Find positions of brackets
    open_bracket_pos = text.find("[")
    close_bracket_pos = text.find("]")

    # Check if brackets are in correct order
    if open_bracket_pos >= close_bracket_pos:
        return text
    # Extract content between brackets
    content_between_brackets = text[open_bracket_pos + 1 : close_bracket_pos]
    # Check if there's a comma between brackets
    if "," not in content_between_brackets:
        return text
    text = text.replace(content_between_brackets, content_between_brackets.replace(", ", "]["))
    text = text.replace(content_between_brackets, content_between_brackets.replace(",", "]["))

    return text


def process_streaming_references_complete(text_buffer: str) -> tuple[str, str]:
    """
    Complete streaming reference processing to ensure reference tags are never split.

    Args:
        text_buffer (str): The accumulated text buffer.

    Returns:
        tuple[str, str]: (processed_text, remaining_buffer)
    """
    import re

    # Pattern to match complete reference tags: [refid:memoriesID]
    complete_pattern = r"\[\d+:[^\]]+\]"

    # Find all complete reference tags
    complete_matches = list(re.finditer(complete_pattern, text_buffer))

    if complete_matches:
        # Find the last complete tag
        last_match = complete_matches[-1]
        end_pos = last_match.end()

        # Check if there's any incomplete reference after the last complete one
        remaining_text = text_buffer[end_pos:]

        # Look for potential incomplete reference patterns after the last complete tag
        incomplete_pattern = r"\[\d*:?[^\]]*$"
        if re.search(incomplete_pattern, remaining_text):
            # There's a potential incomplete reference, find where it starts
            incomplete_match = re.search(incomplete_pattern, remaining_text)
            if incomplete_match:
                incomplete_start = end_pos + incomplete_match.start()
                processed_text = text_buffer[:incomplete_start]
                remaining_buffer = text_buffer[incomplete_start:]

                # Apply reference splitting to the processed text
                processed_text = split_continuous_references(processed_text)
                return processed_text, remaining_buffer

        # No incomplete reference after the last complete tag, process all
        processed_text = split_continuous_references(text_buffer)
        return processed_text, ""

    # Check for incomplete reference tags - be more specific about what constitutes a potential reference
    # Look for opening bracket with number and colon that could be a reference tag
    opening_pattern = r"\[\d+:"
    opening_matches = list(re.finditer(opening_pattern, text_buffer))

    if opening_matches:
        # Find the last opening tag
        last_opening = opening_matches[-1]
        opening_start = last_opening.start()

        # Check if this might be a complete reference tag (has closing bracket after the pattern)
        remaining_text = text_buffer[last_opening.end() :]
        if "]" in remaining_text:
            # This looks like a complete reference tag, process it
            processed_text = split_continuous_references(text_buffer)
            return processed_text, ""
        else:
            # Incomplete reference tag, keep it in buffer
            processed_text = text_buffer[:opening_start]
            processed_text = split_continuous_references(processed_text)
            return processed_text, text_buffer[opening_start:]

    # More sophisticated check for potential reference patterns
    # Only hold back text if we see a pattern that could be the start of a reference tag
    potential_ref_pattern = r"\[\d*:?$"  # Matches [, [1, [12:, etc. at end of buffer
    if re.search(potential_ref_pattern, text_buffer):
        # Find the position of the potential reference start
        match = re.search(potential_ref_pattern, text_buffer)
        if match:
            ref_start = match.start()
            processed_text = text_buffer[:ref_start]
            processed_text = split_continuous_references(processed_text)
            return processed_text, text_buffer[ref_start:]

    # Check for standalone [ only at the very end of the buffer
    # This prevents cutting off mathematical expressions like [ \Delta U = Q - W ]
    if text_buffer.endswith("["):
        # Only hold back the single [ character
        processed_text = text_buffer[:-1]
        processed_text = split_continuous_references(processed_text)
        return processed_text, "["

    # No reference-like patterns found, process all text
    processed_text = split_continuous_references(text_buffer)
    return processed_text, ""


def prepare_reference_data(memories_list: list[TextualMemoryItem]) -> list[dict]:
    # Prepare reference data
    reference = []
    for memories in memories_list:
        if isinstance(memories, TextualMemoryItem):
            memories_json = memories.model_dump()
            memories_json["metadata"]["ref_id"] = f"{memories.id.split('-')[0]}"
            memories_json["metadata"]["embedding"] = []
            memories_json["metadata"]["sources"] = []
            memories_json["metadata"]["memory"] = memories.memory
            memories_json["metadata"]["id"] = memories.id
            reference.append({"metadata": memories_json["metadata"]})
        else:
            memories_json = memories
            memories_json["metadata"]["ref_id"] = f"{memories_json['id'].split('-')[0]}"
            memories_json["metadata"]["embedding"] = []
            memories_json["metadata"]["sources"] = []
            memories_json["metadata"]["memory"] = memories_json["memory"]
            memories_json["metadata"]["id"] = memories_json["id"]
            reference.append({"metadata": memories_json["metadata"]})

    return reference



import os

from abc import ABC

from memos import log
from memos.configs.mem_reader import StrategyStructMemReaderConfig
from memos.configs.parser import ParserConfigFactory
from memos.mem_reader.simple_struct import SimpleStructMemReader, detect_lang
from memos.parsers.factory import ParserFactory
from memos.templates.mem_reader_prompts import (
    SIMPLE_STRUCT_DOC_READER_PROMPT,
    SIMPLE_STRUCT_DOC_READER_PROMPT_ZH,
    SIMPLE_STRUCT_MEM_READER_EXAMPLE,
    SIMPLE_STRUCT_MEM_READER_EXAMPLE_ZH,
)
from memos.templates.mem_reader_strategy_prompts import (
    STRATEGY_STRUCT_MEM_READER_PROMPT,
    STRATEGY_STRUCT_MEM_READER_PROMPT_ZH,
)


logger = log.get_logger(__name__)
STRATEGY_PROMPT_DICT = {
    "chat": {
        "en": STRATEGY_STRUCT_MEM_READER_PROMPT,
        "zh": STRATEGY_STRUCT_MEM_READER_PROMPT_ZH,
        "en_example": SIMPLE_STRUCT_MEM_READER_EXAMPLE,
        "zh_example": SIMPLE_STRUCT_MEM_READER_EXAMPLE_ZH,
    },
    "doc": {"en": SIMPLE_STRUCT_DOC_READER_PROMPT, "zh": SIMPLE_STRUCT_DOC_READER_PROMPT_ZH},
}


class StrategyStructMemReader(SimpleStructMemReader, ABC):
    """Naive implementation of MemReader."""

    def __init__(self, config: StrategyStructMemReaderConfig):
        super().__init__(config)
        self.chat_chunker = config.chat_chunker["config"]

    def _get_llm_response(self, mem_str: str) -> dict:
        lang = detect_lang(mem_str)
        template = STRATEGY_PROMPT_DICT["chat"][lang]
        examples = STRATEGY_PROMPT_DICT["chat"][f"{lang}_example"]
        prompt = template.replace("${conversation}", mem_str)
        if self.config.remove_prompt_example:  # TODO unused
            prompt = prompt.replace(examples, "")
        messages = [{"role": "user", "content": prompt}]
        try:
            response_text = self.llm.generate(messages)
            response_json = self.parse_json_result(response_text)
        except Exception as e:
            logger.error(f"[LLM] Exception during chat generation: {e}")
            response_json = {
                "memory list": [
                    {
                        "key": mem_str[:10],
                        "memory_type": "UserMemory",
                        "value": mem_str,
                        "tags": [],
                    }
                ],
                "summary": mem_str,
            }
        return response_json

    def get_scene_data_info(self, scene_data: list, type: str) -> list[str]:
        """
        Get raw information from scene_data.
        If scene_data contains dictionaries, convert them to strings.
        If scene_data contains file paths, parse them using the parser.

        Args:
            scene_data: List of dialogue information or document paths
            type: Type of scene data: ['doc', 'chat']
        Returns:
            List of strings containing the processed scene data
        """
        results = []

        if type == "chat":
            if self.chat_chunker["chunk_type"] == "content_length":
                content_len_thredshold = self.chat_chunker["chunk_length"]
                for items in scene_data:
                    if not items:
                        continue

                    results.append([])
                    current_length = 0

                    for _i, item in enumerate(items):
                        content_length = (
                            len(item.get("content", ""))
                            if isinstance(item, dict)
                            else len(str(item))
                        )
                        if not results[-1]:
                            results[-1].append(item)
                            current_length = content_length
                            continue

                        if current_length + content_length <= content_len_thredshold:
                            results[-1].append(item)
                            current_length += content_length
                        else:
                            overlap_item = results[-1][-1]
                            overlap_length = (
                                len(overlap_item.get("content", ""))
                                if isinstance(overlap_item, dict)
                                else len(str(overlap_item))
                            )

                            results.append([overlap_item, item])
                            current_length = overlap_length + content_length
            else:
                cut_size, cut_overlap = (
                    self.chat_chunker["chunk_session"],
                    self.chat_chunker["chunk_overlap"],
                )
                for items in scene_data:
                    step = cut_size - cut_overlap
                    end = len(items) - cut_overlap
                    if end <= 0:
                        results.extend([items[:]])
                    else:
                        results.extend([items[i : i + cut_size] for i in range(0, end, step)])

        elif type == "doc":
            parser_config = ParserConfigFactory.model_validate(
                {
                    "backend": "markitdown",
                    "config": {},
                }
            )
            parser = ParserFactory.from_config(parser_config)
            for item in scene_data:
                try:
                    if os.path.exists(item):
                        try:
                            parsed_text = parser.parse(item)
                            results.append({"file": item, "text": parsed_text})
                        except Exception as e:
                            logger.error(f"[SceneParser] Error parsing {item}: {e}")
                            continue
                    else:
                        parsed_text = item
                        results.append({"file": "pure_text", "text": parsed_text})
                except Exception as e:
                    print(f"Error parsing file {item}: {e!s}")

        return results



import asyncio
import time

from datetime import datetime
from typing import Literal

from memos.context.context import ContextThread
from memos.llms.base import BaseLLM
from memos.log import get_logger
from memos.mem_cube.navie import NaiveMemCube
from memos.mem_os.product import _format_mem_block
from memos.mem_reader.base import BaseMemReader
from memos.memories.textual.item import TextualMemoryItem
from memos.templates.mos_prompts import (
    get_memos_prompt,
)
from memos.types import MessageList


logger = get_logger(__name__)


class MOSServer:
    def __init__(
        self,
        mem_reader: BaseMemReader | None = None,
        llm: BaseLLM | None = None,
        online_bot: bool = False,
    ):
        self.mem_reader = mem_reader
        self.chat_llm = llm
        self.online_bot = online_bot

    def chat(
        self,
        query: str,
        user_id: str,
        cube_id: str | None = None,
        mem_cube: NaiveMemCube | None = None,
        history: MessageList | None = None,
        base_prompt: str | None = None,
        internet_search: bool = False,
        moscube: bool = False,
        top_k: int = 10,
        threshold: float = 0.5,
        session_id: str | None = None,
    ) -> str:
        """
        Chat with LLM with memory references and complete response.
        """
        time_start = time.time()
        memories_result = mem_cube.text_mem.search(
            query=query,
            user_name=cube_id,
            top_k=top_k,
            mode="fine",
            manual_close_internet=not internet_search,
            moscube=moscube,
            info={
                "user_id": user_id,
                "session_id": session_id,
                "chat_history": history,
            },
        )

        memories_list = []
        if memories_result:
            memories_list = self._filter_memories_by_threshold(memories_result, threshold)
            new_memories_list = []
            for m in memories_list:
                m.metadata.embedding = []
                new_memories_list.append(m)
            memories_list = new_memories_list
        system_prompt = self._build_system_prompt(memories_list, base_prompt)

        history_info = []
        if history:
            history_info = history[-20:]
        current_messages = [
            {"role": "system", "content": system_prompt},
            *history_info,
            {"role": "user", "content": query},
        ]
        response = self.chat_llm.generate(current_messages)
        time_end = time.time()
        self._start_post_chat_processing(
            user_id=user_id,
            cube_id=cube_id,
            session_id=session_id,
            query=query,
            full_response=response,
            system_prompt=system_prompt,
            time_start=time_start,
            time_end=time_end,
            speed_improvement=0.0,
            current_messages=current_messages,
            mem_cube=mem_cube,
            history=history,
        )
        return response, memories_list

    def add(
        self,
        user_id: str,
        cube_id: str,
        mem_cube: NaiveMemCube,
        messages: MessageList,
        session_id: str | None = None,
        history: MessageList | None = None,
    ) -> list[str]:
        memories = self.mem_reader.get_memory(
            [messages],
            type="chat",
            info={
                "user_id": user_id,
                "session_id": session_id,
                "chat_history": history,
            },
        )
        flattened_memories = [mm for m in memories for mm in m]
        mem_id_list: list[str] = mem_cube.text_mem.add(
            flattened_memories,
            user_name=cube_id,
        )
        return mem_id_list

    def search(
        self,
        user_id: str,
        cube_id: str,
        session_id: str | None = None,
    ) -> None:
        NotImplementedError("Not implemented")

    def _filter_memories_by_threshold(
        self,
        memories: list[TextualMemoryItem],
        threshold: float = 0.30,
        min_num: int = 3,
        memory_type: Literal["OuterMemory"] = "OuterMemory",
    ) -> list[TextualMemoryItem]:
        """
        Filter memories by threshold and type, at least min_num memories for Non-OuterMemory.
        Args:
            memories: list[TextualMemoryItem],
            threshold: float,
            min_num: int,
            memory_type: Literal["OuterMemory"],
        Returns:
            list[TextualMemoryItem]
        """
        sorted_memories = sorted(memories, key=lambda m: m.metadata.relativity, reverse=True)
        filtered_person = [m for m in memories if m.metadata.memory_type != memory_type]
        filtered_outer = [m for m in memories if m.metadata.memory_type == memory_type]
        filtered = []
        per_memory_count = 0
        for m in sorted_memories:
            if m.metadata.relativity >= threshold:
                if m.metadata.memory_type != memory_type:
                    per_memory_count += 1
                filtered.append(m)
        if len(filtered) < min_num:
            filtered = filtered_person[:min_num] + filtered_outer[:min_num]
        else:
            if per_memory_count < min_num:
                filtered += filtered_person[per_memory_count:min_num]
        filtered_memory = sorted(filtered, key=lambda m: m.metadata.relativity, reverse=True)
        return filtered_memory

    def _build_base_system_prompt(
        self,
        base_prompt: str | None = None,
        tone: str = "friendly",
        verbosity: str = "mid",
        mode: str = "enhance",
    ) -> str:
        """
        Build base system prompt without memory references.
        """
        now = datetime.now()
        formatted_date = now.strftime("%Y-%m-%d (%A)")
        sys_body = get_memos_prompt(date=formatted_date, tone=tone, verbosity=verbosity, mode=mode)
        prefix = (base_prompt.strip() + "\n\n") if base_prompt else ""
        return prefix + sys_body

    def _build_system_prompt(
        self,
        memories: list[TextualMemoryItem] | list[str] | None = None,
        base_prompt: str | None = None,
        **kwargs,
    ) -> str:
        """Build system prompt with optional memories context."""
        if base_prompt is None:
            base_prompt = (
                "You are a knowledgeable and helpful AI assistant. "
                "You have access to conversation memories that help you provide more personalized responses. "
                "Use the memories to understand the user's context, preferences, and past interactions. "
                "If memories are provided, reference them naturally when relevant, but don't explicitly mention having memories."
            )

        memory_context = ""
        if memories:
            memory_list = []
            for i, memory in enumerate(memories, 1):
                if isinstance(memory, TextualMemoryItem):
                    text_memory = memory.memory
                else:
                    if not isinstance(memory, str):
                        logger.error("Unexpected memory type.")
                    text_memory = memory
                memory_list.append(f"{i}. {text_memory}")
            memory_context = "\n".join(memory_list)

        if "{memories}" in base_prompt:
            return base_prompt.format(memories=memory_context)
        elif base_prompt and memories:
            # For backward compatibility, append memories if no placeholder is found
            memory_context_with_header = "\n\n## Memories:\n" + memory_context
            return base_prompt + memory_context_with_header
        return base_prompt

    def _build_memory_context(
        self,
        memories_all: list[TextualMemoryItem],
        mode: str = "enhance",
    ) -> str:
        """
        Build memory context to be included in user message.
        """
        if not memories_all:
            return ""

        mem_block_o, mem_block_p = _format_mem_block(memories_all)

        if mode == "enhance":
            return (
                "# Memories\n## PersonalMemory (ordered)\n"
                + mem_block_p
                + "\n## OuterMemory (ordered)\n"
                + mem_block_o
                + "\n\n"
            )
        else:
            mem_block = mem_block_o + "\n" + mem_block_p
            return "# Memories\n## PersonalMemory & OuterMemory (ordered)\n" + mem_block + "\n\n"

    def _extract_references_from_response(self, response: str) -> tuple[str, list[dict]]:
        """
        Extract reference information from the response and return clean text.

        Args:
            response (str): The complete response text.

        Returns:
            tuple[str, list[dict]]: A tuple containing:
                - clean_text: Text with reference markers removed
                - references: List of reference information
        """
        import re

        try:
            references = []
            # Pattern to match [refid:memoriesID]
            pattern = r"\[(\d+):([^\]]+)\]"

            matches = re.findall(pattern, response)
            for ref_number, memory_id in matches:
                references.append({"memory_id": memory_id, "reference_number": int(ref_number)})

            # Remove all reference markers from the text to get clean text
            clean_text = re.sub(pattern, "", response)

            # Clean up any extra whitespace that might be left after removing markers
            clean_text = re.sub(r"\s+", " ", clean_text).strip()

            return clean_text, references
        except Exception as e:
            logger.error(f"Error extracting references from response: {e}", exc_info=True)
            return response, []

    async def _post_chat_processing(
        self,
        user_id: str,
        cube_id: str,
        query: str,
        full_response: str,
        system_prompt: str,
        time_start: float,
        time_end: float,
        speed_improvement: float,
        current_messages: list,
        mem_cube: NaiveMemCube | None = None,
        session_id: str | None = None,
        history: MessageList | None = None,
    ) -> None:
        """
        Asynchronous processing of logs, notifications and memory additions
        """
        try:
            logger.info(
                f"user_id: {user_id}, cube_id: {cube_id}, current_messages: {current_messages}"
            )
            logger.info(f"user_id: {user_id}, cube_id: {cube_id}, full_response: {full_response}")

            clean_response, extracted_references = self._extract_references_from_response(
                full_response
            )
            logger.info(f"Extracted {len(extracted_references)} references from response")

            # Send chat report notifications asynchronously
            if self.online_bot:
                try:
                    from memos.memos_tools.notification_utils import (
                        send_online_bot_notification_async,
                    )

                    # Prepare notification data
                    chat_data = {
                        "query": query,
                        "user_id": user_id,
                        "cube_id": cube_id,
                        "system_prompt": system_prompt,
                        "full_response": full_response,
                    }

                    system_data = {
                        "references": extracted_references,
                        "time_start": time_start,
                        "time_end": time_end,
                        "speed_improvement": speed_improvement,
                    }

                    emoji_config = {"chat": "💬", "system_info": "📊"}

                    await send_online_bot_notification_async(
                        online_bot=self.online_bot,
                        header_name="MemOS Chat Report",
                        sub_title_name="chat_with_references",
                        title_color="#00956D",
                        other_data1=chat_data,
                        other_data2=system_data,
                        emoji=emoji_config,
                    )
                except Exception as e:
                    logger.warning(f"Failed to send chat notification (async): {e}")

            self.add(
                user_id=user_id,
                cube_id=cube_id,
                mem_cube=mem_cube,
                session_id=session_id,
                history=history,
                messages=[
                    {
                        "role": "user",
                        "content": query,
                        "chat_time": str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                    },
                    {
                        "role": "assistant",
                        "content": clean_response,  # Store clean text without reference markers
                        "chat_time": str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                    },
                ],
            )

            logger.info(f"Post-chat processing completed for user {user_id}")

        except Exception as e:
            logger.error(f"Error in post-chat processing for user {user_id}: {e}", exc_info=True)

    def _start_post_chat_processing(
        self,
        user_id: str,
        cube_id: str,
        query: str,
        full_response: str,
        system_prompt: str,
        time_start: float,
        time_end: float,
        speed_improvement: float,
        current_messages: list,
        mem_cube: NaiveMemCube | None = None,
        session_id: str | None = None,
        history: MessageList | None = None,
    ) -> None:
        """
        Asynchronous processing of logs, notifications and memory additions, handle synchronous and asynchronous environments
        """

        def run_async_in_thread():
            """Running asynchronous tasks in a new thread"""
            try:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    loop.run_until_complete(
                        self._post_chat_processing(
                            user_id=user_id,
                            cube_id=cube_id,
                            query=query,
                            full_response=full_response,
                            system_prompt=system_prompt,
                            time_start=time_start,
                            time_end=time_end,
                            speed_improvement=speed_improvement,
                            current_messages=current_messages,
                            mem_cube=mem_cube,
                            session_id=session_id,
                            history=history,
                        )
                    )
                finally:
                    loop.close()
            except Exception as e:
                logger.error(
                    f"Error in thread-based post-chat processing for user {user_id}: {e}",
                    exc_info=True,
                )

        try:
            # Try to get the current event loop
            asyncio.get_running_loop()
            # Create task and store reference to prevent garbage collection
            task = asyncio.create_task(
                self._post_chat_processing(
                    user_id=user_id,
                    cube_id=cube_id,
                    query=query,
                    full_response=full_response,
                    system_prompt=system_prompt,
                    time_start=time_start,
                    time_end=time_end,
                    speed_improvement=speed_improvement,
                    current_messages=current_messages,
                )
            )
            # Add exception handling for the background task
            task.add_done_callback(
                lambda t: logger.error(
                    f"Error in background post-chat processing for user {user_id}: {t.exception()}",
                    exc_info=True,
                )
                if t.exception()
                else None
            )
        except RuntimeError:
            # No event loop, run in a new thread with context propagation
            thread = ContextThread(
                target=run_async_in_thread,
                name=f"PostChatProcessing-{user_id}",
                # Set as a daemon thread to avoid blocking program exit
                daemon=True,
            )
            thread.start()



import json
import os

from datetime import datetime
from typing import Any

from memos.configs.memory import NaiveTextMemoryConfig
from memos.llms.factory import LLMFactory
from memos.log import get_logger
from memos.memories.textual.base import BaseTextMemory
from memos.memories.textual.item import TextualMemoryItem, TextualMemoryMetadata
from memos.types import MessageList


logger = get_logger(__name__)


EXTRACTION_PROMPT_PART_1 = f"""You are a memory extractor. Your task is to extract memories from the given messages.
* You will receive a list of messages, each with a role (user or assistant) and content.
* Your job is to extract the memories from these messages.
* Each memory should be a dictionary with the following keys:
    - "memory": The content of the memory (string). Rephrase the content if necessary.
    - "type": The type of memory (string), e.g., "procedure", "fact", "event", "opinion", etc.
* Current date and time is {datetime.now().isoformat()}.
* Only return the list of memories in JSON format.
* Do not include any other text or explanation.

## Example

### Input

[
    {{"role": "user", "content": "I plan to visit Paris next week."}},
    {{"role": "assistant", "content": "Paris is a beautiful city with many attractions."}},
    {{"role": "user", "content": "I love the Eiffel Tower."}},
    {{"role": "assistant", "content": "The Eiffel Tower is a must-see landmark in Paris."}}
]

### Output

[
    {{"memory": "User plans to visit Paris next week.", "metadata": {{"type": "event"}}}},
    {{"memory": "User loves the Eiffel Tower.", "metadata": {{"type": "opinion"}}}},
]
"""

EXTRACTION_PROMPT_PART_2 = """
## Query

### Input

{messages}

### Output

"""


class NaiveTextMemory(BaseTextMemory):
    """Naive textual memory implementation for storing and retrieving memories."""

    def __init__(self, config: NaiveTextMemoryConfig):
        """Initialize memory with the given configuration."""
        # Set mode from class default or override if needed
        self.mode = getattr(self.__class__, "mode", "sync")
        self.config = config
        self.extractor_llm = LLMFactory.from_config(config.extractor_llm)
        self.memories = []

    def extract(self, messages: MessageList) -> list[TextualMemoryItem]:
        """Extract memories based on the messages."""
        str_messages = json.dumps(messages)
        user_query = EXTRACTION_PROMPT_PART_1 + EXTRACTION_PROMPT_PART_2.format(
            messages=str_messages
        )
        response = self.extractor_llm.generate([{"role": "user", "content": user_query}])
        raw_extracted_memories = json.loads(response)

        # Convert raw dictionaries to TextualMemoryItem objects
        extracted_memories = []
        for memory_dict in raw_extracted_memories:
            # Ensure proper structure with memory and metadata
            memory_content = memory_dict.get("memory", "")
            metadata_dict = memory_dict.get("metadata", {})

            # Create a TextualMemoryItem with properly structured metadata
            memory_item = TextualMemoryItem(memory=memory_content, metadata=metadata_dict)
            extracted_memories.append(memory_item)

        return extracted_memories

    def add(self, memories: list[TextualMemoryItem | dict[str, Any]]) -> None:
        """Add memories."""
        for m in memories:
            # Convert dict to TextualMemoryItem if needed
            memory_item = TextualMemoryItem(**m) if isinstance(m, dict) else m

            # Convert to dictionary for storage
            memory_dict = memory_item.model_dump()

            if memory_dict["id"] not in [m["id"] for m in self.memories]:
                self.memories.append(memory_dict)

    def update(self, memory_id: str, new_memory: TextualMemoryItem | dict[str, Any]) -> None:
        """Update a memory by memory_id."""
        # Convert dict to TextualMemoryItem if needed
        memory_item = (
            TextualMemoryItem(**new_memory) if isinstance(new_memory, dict) else new_memory
        )

        # Ensure the memory item has the correct ID
        memory_item.id = memory_id
        memory_dict = memory_item.model_dump()

        for i, memory in enumerate(self.memories):
            if memory["id"] == memory_id:
                self.memories[i] = memory_dict
                break

    def search(self, query: str, top_k: int, **kwargs) -> list[TextualMemoryItem]:
        """Search for memories based on a query."""
        sims = [
            (memory, len(set(query.split()) & set(memory["memory"].split())))
            for memory in self.memories
        ]
        sims.sort(key=lambda x: x[1], reverse=True)
        # Convert search results to TextualMemoryItem objects
        return [TextualMemoryItem(**memory) for memory, _ in sims[:top_k]]

    def get(self, memory_id: str) -> TextualMemoryItem:
        """Get a memory by its ID."""
        for memory in self.memories:
            if memory["id"] == memory_id:
                return TextualMemoryItem(**memory)
        # Return empty memory item if not found
        return TextualMemoryItem(id=memory_id, memory="", metadata=TextualMemoryMetadata())

    def get_all(self) -> list[TextualMemoryItem]:
        """Get all memories."""
        return [TextualMemoryItem(**memory) for memory in self.memories]

    def get_by_ids(self, memory_ids: list[str]) -> list[TextualMemoryItem]:
        """Get memories by their IDs.
        Args:
            memory_ids (list[str]): List of memory IDs to retrieve.
        Returns:
            list[TextualMemoryItem]: List of memories with the specified IDs.
        """
        return [self.get(memory_id) for memory_id in memory_ids]

    def delete(self, memory_ids: list[str]) -> None:
        """Delete memories.
        Args:
            memory_ids (list[str]): List of memory IDs to delete.
        """
        self.memories = [m for m in self.memories if m["id"] not in memory_ids]

    def delete_all(self) -> None:
        """Delete all memories."""
        self.memories = []

    def load(self, dir: str) -> None:
        try:
            with open(os.path.join(dir, self.config.memory_filename), encoding="utf-8") as file:
                raw_memories = json.load(file)
                self.add(raw_memories)
        except FileNotFoundError:
            logger.error(f"Directory not found: {dir}")
        except json.JSONDecodeError:
            logger.error(f"Error decoding JSON from file in directory: {dir}")
        except Exception as e:
            logger.error(f"An error occurred while loading memories: {e}")

    def dump(self, dir: str) -> None:
        try:
            os.makedirs(dir, exist_ok=True)
            memory_file = os.path.join(dir, self.config.memory_filename)
            with open(memory_file, "w", encoding="utf-8") as file:
                json.dump(self.memories, file, indent=4, ensure_ascii=False)
        except Exception as e:
            logger.error(f"An error occurred while dumping memories: {e}")
            raise

    def drop(
        self,
    ) -> None:
        pass



"""
Thread-safe dictionary wrapper for concurrent access with optimized read-write locks.
"""

import threading

from collections.abc import ItemsView, Iterator, KeysView, ValuesView
from typing import Generic, TypeVar

from memos.log import get_logger
from memos.utils import timed


K = TypeVar("K")
V = TypeVar("V")

logger = get_logger(__name__)


class ReadWriteLock:
    """A simple read-write lock implementation. use for product-server scenario"""

    def __init__(self):
        self._read_ready = threading.Condition(threading.RLock())
        self._readers = 0

    @timed
    def acquire_read(self):
        """Acquire a read lock. Multiple readers can hold the lock simultaneously."""
        self._read_ready.acquire()
        try:
            self._readers += 1
        finally:
            self._read_ready.release()

    def release_read(self):
        """Release a read lock."""
        self._read_ready.acquire()
        try:
            self._readers -= 1
            if self._readers == 0:
                self._read_ready.notify_all()
        finally:
            self._read_ready.release()

    @timed
    def acquire_write(self):
        """Acquire a write lock. Only one writer can hold the lock."""
        self._read_ready.acquire()
        while self._readers > 0:
            self._read_ready.wait()

    def release_write(self):
        """Release a write lock."""
        self._read_ready.release()


class ThreadSafeDict(Generic[K, V]):
    """
    A thread-safe dictionary wrapper with optimized read-write locks.

    This class allows multiple concurrent readers while ensuring exclusive access for writers.
    Read operations (get, contains, iteration) can happen concurrently.
    Write operations (set, delete, update) are exclusive.
    """

    def __init__(self, initial_dict: dict[K, V] | None = None):
        """
        Initialize the thread-safe dictionary.

        Args:
            initial_dict: Optional initial dictionary to copy from
        """
        self._dict: dict[K, V] = initial_dict.copy() if initial_dict else {}
        self._lock = ReadWriteLock()

    @timed
    def __getitem__(self, key: K) -> V:
        """Get item by key."""
        self._lock.acquire_read()
        try:
            return self._dict[key]
        finally:
            self._lock.release_read()

    @timed
    def __setitem__(self, key: K, value: V) -> None:
        """Set item by key."""
        self._lock.acquire_write()
        try:
            self._dict[key] = value
        finally:
            self._lock.release_write()

    @timed
    def __delitem__(self, key: K) -> None:
        """Delete item by key."""
        self._lock.acquire_write()
        try:
            del self._dict[key]
        finally:
            self._lock.release_write()

    @timed
    def __contains__(self, key: K) -> bool:
        """Check if key exists in dictionary."""
        self._lock.acquire_read()
        try:
            return key in self._dict
        finally:
            self._lock.release_read()

    @timed
    def __len__(self) -> int:
        """Get length of dictionary."""
        self._lock.acquire_read()
        try:
            return len(self._dict)
        finally:
            self._lock.release_read()

    def __bool__(self) -> bool:
        """Check if dictionary is not empty."""
        self._lock.acquire_read()
        try:
            return bool(self._dict)
        finally:
            self._lock.release_read()

    @timed
    def __iter__(self) -> Iterator[K]:
        """Iterate over keys. Returns a snapshot to avoid iteration issues."""
        self._lock.acquire_read()
        try:
            # Return a snapshot of keys to avoid iteration issues
            return iter(list(self._dict.keys()))
        finally:
            self._lock.release_read()

    @timed
    def get(self, key: K, default: V | None = None) -> V:
        """Get item by key with optional default."""
        self._lock.acquire_read()
        try:
            return self._dict.get(key, default)
        finally:
            self._lock.release_read()

    @timed
    def pop(self, key: K, *args) -> V:
        """Pop item by key."""
        self._lock.acquire_write()
        try:
            return self._dict.pop(key, *args)
        finally:
            self._lock.release_write()

    @timed
    def update(self, *args, **kwargs) -> None:
        """Update dictionary."""
        self._lock.acquire_write()
        try:
            self._dict.update(*args, **kwargs)
        finally:
            self._lock.release_write()

    @timed
    def clear(self) -> None:
        """Clear all items."""
        self._lock.acquire_write()
        try:
            self._dict.clear()
        finally:
            self._lock.release_write()

    @timed
    def keys(self) -> KeysView[K]:
        """Get dictionary keys view (snapshot)."""
        self._lock.acquire_read()
        try:
            return list(self._dict.keys())
        finally:
            self._lock.release_read()

    @timed
    def values(self) -> ValuesView[V]:
        """Get dictionary values view (snapshot)."""
        self._lock.acquire_read()
        try:
            return list(self._dict.values())
        finally:
            self._lock.release_read()

    @timed
    def items(self) -> ItemsView[K, V]:
        """Get dictionary items view (snapshot)."""
        self._lock.acquire_read()
        try:
            return list(self._dict.items())
        finally:
            self._lock.release_read()

    @timed
    def copy(self) -> dict[K, V]:
        """Create a copy of the dictionary."""
        self._lock.acquire_read()
        try:
            return self._dict.copy()
        finally:
            self._lock.release_read()

    @timed
    def setdefault(self, key: K, default: V | None = None) -> V:
        """Set default value for key if not exists."""
        self._lock.acquire_write()
        try:
            return self._dict.setdefault(key, default)
        finally:
            self._lock.release_write()

    def __repr__(self) -> str:
        """String representation."""
        self._lock.acquire_read()
        try:
            return f"ThreadSafeDict({self._dict})"
        finally:
            self._lock.release_read()

    def __str__(self) -> str:
        """String representation."""
        self._lock.acquire_read()
        try:
            return str(self._dict)
        finally:
            self._lock.release_read()


class SimpleThreadSafeDict(Generic[K, V]):
    """
    Simple thread-safe dictionary with exclusive locks for all operations.
    Use this if you prefer simplicity over performance.
    """

    def __init__(self, initial_dict: dict[K, V] | None = None):
        self._dict: dict[K, V] = initial_dict.copy() if initial_dict else {}
        self._lock = threading.RLock()

    def __getitem__(self, key: K) -> V:
        with self._lock:
            return self._dict[key]

    def __setitem__(self, key: K, value: V) -> None:
        with self._lock:
            self._dict[key] = value

    def __delitem__(self, key: K) -> None:
        with self._lock:
            del self._dict[key]

    def __contains__(self, key: K) -> bool:
        with self._lock:
            return key in self._dict

    def __len__(self) -> int:
        with self._lock:
            return len(self._dict)

    def __bool__(self) -> bool:
        with self._lock:
            return bool(self._dict)

    def __iter__(self) -> Iterator[K]:
        with self._lock:
            return iter(list(self._dict.keys()))

    def get(self, key: K, default: V | None = None) -> V:
        with self._lock:
            return self._dict.get(key, default)

    def pop(self, key: K, *args) -> V:
        with self._lock:
            return self._dict.pop(key, *args)

    def update(self, *args, **kwargs) -> None:
        with self._lock:
            self._dict.update(*args, **kwargs)

    def clear(self) -> None:
        with self._lock:
            self._dict.clear()

    def keys(self):
        with self._lock:
            return list(self._dict.keys())

    def values(self):
        with self._lock:
            return list(self._dict.values())

    def items(self):
        with self._lock:
            return list(self._dict.items())

    def copy(self) -> dict[K, V]:
        with self._lock:
            return self._dict.copy()

    def setdefault(self, key: K, default: V | None = None) -> V:
        with self._lock:
            return self._dict.setdefault(key, default)



from typing import TypeVar

from sqlalchemy import Index
from sqlalchemy.engine import Engine

from memos.log import get_logger
from memos.mem_scheduler.schemas.monitor_schemas import (
    MemoryMonitorItem,
    MemoryMonitorManager,
    QueryMonitorItem,
    QueryMonitorQueue,
)

from .base_model import BaseDBManager, LockableORM


logger = get_logger(__name__)

# Type variables for generic type hints
T = TypeVar("T")  # The model type (MemoryMonitorManager, QueryMonitorManager, etc.)
ORM = TypeVar("ORM")  # The ORM model type


class MemoryMonitorManagerORM(LockableORM):
    """ORM model for MemoryMonitorManager persistence

    This table stores serialized MemoryMonitorManager instances with
    proper indexing for efficient user and memory cube lookups.
    """

    __tablename__ = "memory_monitor_manager"

    # Database indexes for performance optimization
    __table_args__ = (Index("idx_memory_monitor_user_memcube", "user_id", "mem_cube_id"),)


class QueryMonitorQueueORM(LockableORM):
    """ORM model for QueryMonitorQueue persistence

    This table stores serialized QueryMonitorQueue instances with
    proper indexing for efficient user and memory cube lookups.
    """

    __tablename__ = "query_monitor_queue"

    # Database indexes for performance optimization
    __table_args__ = (Index("idx_query_monitor_user_memcube", "user_id", "mem_cube_id"),)


class DBManagerForMemoryMonitorManager(BaseDBManager):
    """Database manager for MemoryMonitorManager objects

    This class handles persistence, synchronization, and locking
    for MemoryMonitorManager instances in the database.
    """

    def __init__(
        self,
        engine: Engine,
        user_id: str | None = None,
        mem_cube_id: str | None = None,
        obj: MemoryMonitorManager | None = None,
        lock_timeout: int = 10,
    ):
        """
        Initialize the MemoryMonitorManager database manager.

        Args:
            engine: SQLAlchemy engine instance
            user_id: Unique identifier for the user
            mem_cube_id: Unique identifier for the memory cube
            obj: Optional MemoryMonitorManager instance to manage
            lock_timeout: Timeout in seconds for lock acquisition
        """
        super().__init__(
            engine=engine, user_id=user_id, mem_cube_id=mem_cube_id, lock_timeout=lock_timeout
        )
        self.obj: MemoryMonitorManager | None = obj

    @property
    def orm_class(self) -> type[MemoryMonitorManagerORM]:
        return MemoryMonitorManagerORM

    @property
    def obj_class(self) -> type[MemoryMonitorManager]:
        return MemoryMonitorManager

    def merge_items(
        self,
        orm_instance: MemoryMonitorManagerORM,
        obj_instance: MemoryMonitorManager,
        size_limit: int,
    ):
        """Merge memory monitor items from database with current object

        This method combines items from the database with items in the current
        object, prioritizing current object items and applying size limits.

        Args:
            orm_instance: ORM instance containing serialized database data
            obj_instance: Current MemoryMonitorManager instance
            size_limit: Maximum number of items to keep after merge

        Returns:
            Updated obj_instance with merged items
        """
        logger.debug(f"Starting merge_items for MemoryMonitorManager with size_limit={size_limit}")

        try:
            # Deserialize the database instance
            db_instance: MemoryMonitorManager = MemoryMonitorManager.from_json(
                orm_instance.serialized_data
            )
        except Exception as e:
            logger.error(f"Failed to deserialize database instance: {e}", exc_info=True)
            logger.warning("Skipping merge due to deserialization error, using current object only")
            return obj_instance

        # Merge items - prioritize existing ones in current object
        merged_items: list[MemoryMonitorItem] = []
        seen_ids = set()

        # First, add all items from current object (higher priority)
        for item in obj_instance.memories:
            if item.item_id not in seen_ids:
                merged_items.append(item)
                seen_ids.add(item.item_id)

        # Then, add items from database that aren't in current object
        for item in db_instance.memories:
            if item.item_id not in seen_ids:
                merged_items.append(item)
                seen_ids.add(item.item_id)

        # Apply size limit if specified (keep most recent items)
        if size_limit is not None and size_limit > 0:
            try:
                # Sort by sorting_score descending (highest priority first) and take top N
                # Note: MemoryMonitorItem doesn't have timestamp, so we use sorting_score instead
                merged_items = sorted(merged_items, key=lambda x: x.sorting_score, reverse=True)[
                    :size_limit
                ]
                logger.debug(f"Applied size limit of {size_limit}, kept {len(merged_items)} items")
            except AttributeError as e:
                logger.error(f"Error sorting MemoryMonitorItem objects: {e}")
                logger.error(
                    "Available attributes: "
                    + ", ".join(dir(merged_items[0]) if merged_items else [])
                )
                raise
            except Exception as e:
                logger.error(f"Unexpected error during sorting: {e}")
                raise

        # Update the object with merged items
        obj_instance.memories = merged_items

        logger.info(
            f"Merged {len(merged_items)} memory items for {obj_instance} (size_limit: {size_limit})"
        )

        return obj_instance


class DBManagerForQueryMonitorQueue(BaseDBManager):
    """Database manager for QueryMonitorQueue objects

    This class handles persistence, synchronization, and locking
    for QueryMonitorQueue instances in the database.
    """

    def __init__(
        self,
        engine: Engine,
        user_id: str | None = None,
        mem_cube_id: str | None = None,
        obj: QueryMonitorQueue | None = None,
        lock_timeout: int = 10,
    ):
        """
        Initialize the QueryMonitorQueue database manager.

        Args:
            engine: SQLAlchemy engine instance
            user_id: Unique identifier for the user
            mem_cube_id: Unique identifier for the memory cube
            obj: Optional QueryMonitorQueue instance to manage
            lock_timeout: Timeout in seconds for lock acquisition
        """
        super().__init__(
            engine=engine, user_id=user_id, mem_cube_id=mem_cube_id, lock_timeout=lock_timeout
        )
        self.obj: QueryMonitorQueue | None = obj

    @property
    def orm_class(self) -> type[QueryMonitorQueueORM]:
        return QueryMonitorQueueORM

    @property
    def obj_class(self) -> type[QueryMonitorQueue]:
        return QueryMonitorQueue

    def merge_items(
        self, orm_instance: QueryMonitorQueueORM, obj_instance: QueryMonitorQueue, size_limit: int
    ):
        """Merge query monitor items from database with current queue

        This method combines items from the database with items in the current
        queue, prioritizing current queue items and applying size limits.

        Args:
            orm_instance: ORM instance containing serialized database data
            obj_instance: Current QueryMonitorQueue instance
            size_limit: Maximum number of items to keep after merge

        Returns:
            Updated obj_instance with merged items
        """
        try:
            # Deserialize the database instance
            db_instance: QueryMonitorQueue = QueryMonitorQueue.from_json(
                orm_instance.serialized_data
            )
        except Exception as e:
            logger.error(f"Failed to deserialize database instance: {e}")
            logger.warning("Skipping merge due to deserialization error, using current object only")
            return obj_instance

        # Merge items - prioritize existing ones in current object
        merged_items: list[QueryMonitorItem] = []
        seen_ids = set()

        # First, add all items from current queue (higher priority)
        for item in obj_instance.get_queue_content_without_pop():
            if item.item_id not in seen_ids:
                merged_items.append(item)
                seen_ids.add(item.item_id)

        # Then, add items from database queue that aren't in current queue
        for item in db_instance.get_queue_content_without_pop():
            if item.item_id not in seen_ids:
                merged_items.append(item)
                seen_ids.add(item.item_id)

        # Apply size limit if specified (keep most recent items)
        if size_limit is not None and size_limit > 0:
            # Sort by timestamp descending (newest first) and take top N
            merged_items = sorted(merged_items, key=lambda x: x.timestamp, reverse=True)[
                :size_limit
            ]

        # Update the queue with merged items
        obj_instance.clear()  # Clear existing items
        for item in merged_items:
            obj_instance.put(item)  # Add merged items back

        logger.info(
            f"Merged {len(merged_items)} query items for {obj_instance} (size_limit: {size_limit})"
        )

        return obj_instance



import logging
import os

from typing import Any, Generic, TypeVar

from dotenv import load_dotenv
from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.responses import JSONResponse, RedirectResponse
from pydantic import BaseModel, Field

from memos.api.middleware.request_context import RequestContextMiddleware
from memos.configs.mem_os import MOSConfig
from memos.mem_os.main import MOS
from memos.mem_user.user_manager import UserManager, UserRole


# Configure logging
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

# Load environment variables
load_dotenv()

T = TypeVar("T")

# Default configuration
DEFAULT_CONFIG = {
    "user_id": os.getenv("MOS_USER_ID", "default_user"),
    "session_id": os.getenv("MOS_SESSION_ID", "default_session"),
    "enable_textual_memory": True,
    "enable_activation_memory": False,
    "top_k": int(os.getenv("MOS_TOP_K", "5")),
    "chat_model": {
        "backend": os.getenv("MOS_CHAT_MODEL_PROVIDER", "openai"),
        "config": {
            "model_name_or_path": os.getenv("MOS_CHAT_MODEL", "gpt-3.5-turbo"),
            "api_key": os.getenv("OPENAI_API_KEY", "apikey"),
            "temperature": float(os.getenv("MOS_CHAT_TEMPERATURE", "0.7")),
            "api_base": os.getenv("OPENAI_API_BASE", "https://api.openai.com/v1"),
        },
    },
}

# Initialize MOS instance with lazy initialization
MOS_INSTANCE = None


def get_mos_instance():
    """Get or create MOS instance with default user creation."""
    global MOS_INSTANCE
    if MOS_INSTANCE is None:
        # Create a temporary MOS instance to access user manager
        temp_config = MOSConfig(**DEFAULT_CONFIG)
        temp_mos = MOS.__new__(MOS)
        temp_mos.config = temp_config
        temp_mos.user_id = temp_config.user_id
        temp_mos.session_id = temp_config.session_id
        temp_mos.mem_cubes = {}
        temp_mos.chat_llm = None  # Will be initialized later
        temp_mos.user_manager = UserManager()

        # Create default user if it doesn't exist
        if not temp_mos.user_manager.validate_user(temp_config.user_id):
            temp_mos.user_manager.create_user(
                user_name=temp_config.user_id, role=UserRole.USER, user_id=temp_config.user_id
            )
            logger.info(f"Created default user: {temp_config.user_id}")

        # Now create the actual MOS instance
        MOS_INSTANCE = MOS(config=temp_config)

    return MOS_INSTANCE


app = FastAPI(
    title="MemOS REST APIs",
    description="A REST API for managing and searching memories using MemOS.",
    version="1.0.0",
)

app.add_middleware(RequestContextMiddleware)


class BaseRequest(BaseModel):
    """Base model for all requests."""

    user_id: str | None = Field(
        None, description="User ID for the request", json_schema_extra={"example": "user123"}
    )


class BaseResponse(BaseModel, Generic[T]):
    """Base model for all responses."""

    code: int = Field(200, description="Response status code", json_schema_extra={"example": 200})
    message: str = Field(
        ..., description="Response message", json_schema_extra={"example": "Operation successful"}
    )
    data: T | None = Field(None, description="Response data")


class Message(BaseModel):
    role: str = Field(
        ...,
        description="Role of the message (user or assistant).",
        json_schema_extra={"example": "user"},
    )
    content: str = Field(
        ...,
        description="Message content.",
        json_schema_extra={"example": "Hello, how can I help you?"},
    )


class MemoryCreate(BaseRequest):
    messages: list[Message] | None = Field(
        None,
        description="List of messages to store.",
        json_schema_extra={"example": [{"role": "user", "content": "Hello"}]},
    )
    mem_cube_id: str | None = Field(
        None, description="ID of the memory cube", json_schema_extra={"example": "cube123"}
    )
    memory_content: str | None = Field(
        None,
        description="Content to store as memory",
        json_schema_extra={"example": "This is a memory content"},
    )
    doc_path: str | None = Field(
        None,
        description="Path to document to store",
        json_schema_extra={"example": "/path/to/document.txt"},
    )


class SearchRequest(BaseRequest):
    query: str = Field(
        ...,
        description="Search query.",
        json_schema_extra={"example": "How to implement a feature?"},
    )
    install_cube_ids: list[str] | None = Field(
        None,
        description="List of cube IDs to search in",
        json_schema_extra={"example": ["cube123", "cube456"]},
    )


class MemCubeRegister(BaseRequest):
    mem_cube_name_or_path: str = Field(
        ...,
        description="Name or path of the MemCube to register.",
        json_schema_extra={"example": "/path/to/cube"},
    )
    mem_cube_id: str | None = Field(
        None, description="ID for the MemCube", json_schema_extra={"example": "cube123"}
    )


class ChatRequest(BaseRequest):
    query: str = Field(
        ...,
        description="Chat query message.",
        json_schema_extra={"example": "What is the latest update?"},
    )


class UserCreate(BaseRequest):
    user_name: str | None = Field(
        None, description="Name of the user", json_schema_extra={"example": "john_doe"}
    )
    role: str = Field("user", description="Role of the user", json_schema_extra={"example": "user"})
    user_id: str = Field(..., description="User ID", json_schema_extra={"example": "user123"})


class CubeShare(BaseRequest):
    target_user_id: str = Field(
        ..., description="Target user ID to share with", json_schema_extra={"example": "user456"}
    )


class SimpleResponse(BaseResponse[None]):
    """Simple response model for operations without data return."""


class ConfigResponse(BaseResponse[None]):
    """Response model for configuration endpoint."""


class MemoryResponse(BaseResponse[dict]):
    """Response model for memory operations."""


class SearchResponse(BaseResponse[dict]):
    """Response model for search operations."""


class ChatResponse(BaseResponse[str]):
    """Response model for chat operations."""


class UserResponse(BaseResponse[dict]):
    """Response model for user operations."""


class UserListResponse(BaseResponse[list]):
    """Response model for user list operations."""


@app.post("/configure", summary="Configure MemOS", response_model=ConfigResponse)
async def set_config(config: MOSConfig):
    """Set MemOS configuration."""
    global MOS_INSTANCE

    # Create a temporary user manager to check/create default user
    temp_user_manager = UserManager()

    # Create default user if it doesn't exist
    if not temp_user_manager.validate_user(config.user_id):
        temp_user_manager.create_user(
            user_name=config.user_id, role=UserRole.USER, user_id=config.user_id
        )
        logger.info(f"Created default user: {config.user_id}")

    # Now create the MOS instance
    MOS_INSTANCE = MOS(config=config)
    return ConfigResponse(message="Configuration set successfully")


@app.post("/users", summary="Create a new user", response_model=UserResponse)
async def create_user(user_create: UserCreate):
    """Create a new user."""
    mos_instance = get_mos_instance()
    role = UserRole(user_create.role)
    user_id = mos_instance.create_user(
        user_id=user_create.user_id, role=role, user_name=user_create.user_name
    )
    return UserResponse(message="User created successfully", data={"user_id": user_id})


@app.get("/users", summary="List all users", response_model=UserListResponse)
async def list_users():
    """List all active users."""
    mos_instance = get_mos_instance()
    users = mos_instance.list_users()
    return UserListResponse(message="Users retrieved successfully", data=users)


@app.get("/users/me", summary="Get current user info", response_model=UserResponse)
async def get_user_info():
    """Get current user information including accessible cubes."""
    mos_instance = get_mos_instance()
    user_info = mos_instance.get_user_info()
    return UserResponse(message="User info retrieved successfully", data=user_info)


@app.post("/mem_cubes", summary="Register a MemCube", response_model=SimpleResponse)
async def register_mem_cube(mem_cube: MemCubeRegister):
    """Register a new MemCube."""
    mos_instance = get_mos_instance()
    mos_instance.register_mem_cube(
        mem_cube_name_or_path=mem_cube.mem_cube_name_or_path,
        mem_cube_id=mem_cube.mem_cube_id,
        user_id=mem_cube.user_id,
    )
    return SimpleResponse(message="MemCube registered successfully")


@app.delete(
    "/mem_cubes/{mem_cube_id}", summary="Unregister a MemCube", response_model=SimpleResponse
)
async def unregister_mem_cube(mem_cube_id: str, user_id: str | None = None):
    """Unregister a MemCube."""
    mos_instance = get_mos_instance()
    mos_instance.unregister_mem_cube(mem_cube_id=mem_cube_id, user_id=user_id)
    return SimpleResponse(message="MemCube unregistered successfully")


@app.post(
    "/mem_cubes/{cube_id}/share",
    summary="Share a cube with another user",
    response_model=SimpleResponse,
)
async def share_cube(cube_id: str, share_request: CubeShare):
    """Share a cube with another user."""
    mos_instance = get_mos_instance()
    success = mos_instance.share_cube_with_user(cube_id, share_request.target_user_id)
    if success:
        return SimpleResponse(message="Cube shared successfully")
    else:
        raise ValueError("Failed to share cube")


@app.post("/memories", summary="Create memories", response_model=SimpleResponse)
async def add_memory(memory_create: MemoryCreate):
    """Store new memories in a MemCube."""
    if not any([memory_create.messages, memory_create.memory_content, memory_create.doc_path]):
        raise ValueError("Either messages, memory_content, or doc_path must be provided")
    mos_instance = get_mos_instance()
    if memory_create.messages:
        messages = [m.model_dump() for m in memory_create.messages]
        mos_instance.add(
            messages=messages,
            mem_cube_id=memory_create.mem_cube_id,
            user_id=memory_create.user_id,
        )
    elif memory_create.memory_content:
        mos_instance.add(
            memory_content=memory_create.memory_content,
            mem_cube_id=memory_create.mem_cube_id,
            user_id=memory_create.user_id,
        )
    elif memory_create.doc_path:
        mos_instance.add(
            doc_path=memory_create.doc_path,
            mem_cube_id=memory_create.mem_cube_id,
            user_id=memory_create.user_id,
        )
    return SimpleResponse(message="Memories added successfully")


@app.get("/memories", summary="Get all memories", response_model=MemoryResponse)
async def get_all_memories(
    mem_cube_id: str | None = None,
    user_id: str | None = None,
):
    """Retrieve all memories from a MemCube."""
    mos_instance = get_mos_instance()
    result = mos_instance.get_all(mem_cube_id=mem_cube_id, user_id=user_id)
    return MemoryResponse(message="Memories retrieved successfully", data=result)


@app.get(
    "/memories/{mem_cube_id}/{memory_id}", summary="Get a memory", response_model=MemoryResponse
)
async def get_memory(mem_cube_id: str, memory_id: str, user_id: str | None = None):
    """Retrieve a specific memory by ID from a MemCube."""
    mos_instance = get_mos_instance()
    result = mos_instance.get(mem_cube_id=mem_cube_id, memory_id=memory_id, user_id=user_id)
    return MemoryResponse(message="Memory retrieved successfully", data=result)


@app.post("/search", summary="Search memories", response_model=SearchResponse)
async def search_memories(search_req: SearchRequest):
    """Search for memories across MemCubes."""
    mos_instance = get_mos_instance()
    result = mos_instance.search(
        query=search_req.query,
        user_id=search_req.user_id,
        install_cube_ids=search_req.install_cube_ids,
    )
    return SearchResponse(message="Search completed successfully", data=result)


@app.put(
    "/memories/{mem_cube_id}/{memory_id}", summary="Update a memory", response_model=SimpleResponse
)
async def update_memory(
    mem_cube_id: str, memory_id: str, updated_memory: dict[str, Any], user_id: str | None = None
):
    """Update an existing memory in a MemCube."""
    mos_instance = get_mos_instance()
    mos_instance.update(
        mem_cube_id=mem_cube_id,
        memory_id=memory_id,
        text_memory_item=updated_memory,
        user_id=user_id,
    )
    return SimpleResponse(message="Memory updated successfully")


@app.delete(
    "/memories/{mem_cube_id}/{memory_id}", summary="Delete a memory", response_model=SimpleResponse
)
async def delete_memory(mem_cube_id: str, memory_id: str, user_id: str | None = None):
    """Delete a specific memory from a MemCube."""
    mos_instance = get_mos_instance()
    mos_instance.delete(mem_cube_id=mem_cube_id, memory_id=memory_id, user_id=user_id)
    return SimpleResponse(message="Memory deleted successfully")


@app.delete("/memories/{mem_cube_id}", summary="Delete all memories", response_model=SimpleResponse)
async def delete_all_memories(mem_cube_id: str, user_id: str | None = None):
    """Delete all memories from a MemCube."""
    mos_instance = get_mos_instance()
    mos_instance.delete_all(mem_cube_id=mem_cube_id, user_id=user_id)
    return SimpleResponse(message="All memories deleted successfully")


@app.post("/chat", summary="Chat with MemOS", response_model=ChatResponse)
async def chat(chat_req: ChatRequest):
    """Chat with the MemOS system."""
    mos_instance = get_mos_instance()
    response = mos_instance.chat(query=chat_req.query, user_id=chat_req.user_id)
    if response is None:
        raise ValueError("No response generated")
    return ChatResponse(message="Chat response generated", data=response)


@app.get("/", summary="Redirect to the OpenAPI documentation", include_in_schema=False)
async def home():
    """Redirect to the OpenAPI documentation."""
    return RedirectResponse(url="/docs", status_code=307)


@app.exception_handler(ValueError)
async def value_error_handler(request: Request, exc: ValueError):
    """Handle ValueError exceptions globally."""
    return JSONResponse(
        status_code=400,
        content={"code": 400, "message": str(exc), "data": None},
    )


@app.exception_handler(Exception)
async def global_exception_handler(request: Request, exc: Exception):
    """Handle all unhandled exceptions globally."""
    logger.exception("Unhandled error:")
    return JSONResponse(
        status_code=500,
        content={"code": 500, "message": str(exc), "data": None},
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=8000, help="Port to run the server on")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Host to run the server on")
    parser.add_argument("--reload", action="store_true", help="Enable auto-reload for development")
    args = parser.parse_args()



# memos/configs/reranker.py
from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field


class RerankerConfigFactory(BaseModel):
    """
    {
      "backend": "http_bge" | "cosine_local" | "noop",
      "config": { ... backend-specific ... }
    }
    """

    backend: str = Field(..., description="Reranker backend id")
    config: dict[str, Any] = Field(default_factory=dict, description="Backend-specific options")



import json
import ssl
import threading
import time

from pathlib import Path

from memos.configs.mem_scheduler import AuthConfig, RabbitMQConfig
from memos.context.context import ContextThread
from memos.dependency import require_python_package
from memos.log import get_logger
from memos.mem_scheduler.general_modules.base import BaseSchedulerModule
from memos.mem_scheduler.general_modules.misc import AutoDroppingQueue
from memos.mem_scheduler.schemas.general_schemas import DIRECT_EXCHANGE_TYPE, FANOUT_EXCHANGE_TYPE


logger = get_logger(__name__)


class RabbitMQSchedulerModule(BaseSchedulerModule):
    @require_python_package(
        import_name="pika",
        install_command="pip install pika",
        install_link="https://pika.readthedocs.io/en/stable/index.html",
    )
    def __init__(self):
        """
        Initialize RabbitMQ connection settings.
        """
        super().__init__()

        # RabbitMQ settings
        self.rabbitmq_config: RabbitMQConfig | None = None
        self.rabbit_queue_name = "memos-scheduler"
        self.rabbitmq_exchange_name = "memos-fanout"
        self.rabbitmq_exchange_type = FANOUT_EXCHANGE_TYPE
        self.rabbitmq_connection = None
        self.rabbitmq_channel = None

        # fixed params
        self.rabbitmq_message_cache_max_size = 10  # Max 10 messages
        self.rabbitmq_message_cache = AutoDroppingQueue(
            maxsize=self.rabbitmq_message_cache_max_size
        )
        self.rabbitmq_connection_attempts = 3  # Max retry attempts on connection failure
        self.rabbitmq_retry_delay = 5  # Delay (seconds) between retries
        self.rabbitmq_heartbeat = 60  # Heartbeat interval (seconds) for connectio
        self.rabbitmq_conn_max_waiting_seconds = 30
        self.rabbitmq_conn_sleep_seconds = 1

        # Thread management
        self._rabbitmq_io_loop_thread = None  # For IOLoop execution
        self._rabbitmq_stop_flag = False  # Graceful shutdown flag
        self._rabbitmq_lock = threading.Lock()  # Ensure thread safety

    def is_rabbitmq_connected(self) -> bool:
        """Check if RabbitMQ connection is alive"""
        return (
            self.rabbitmq_connection
            and self.rabbitmq_connection.is_open
            and self.rabbitmq_channel
            and self.rabbitmq_channel.is_open
        )

    def initialize_rabbitmq(
        self, config: dict | None | RabbitMQConfig = None, config_path: str | Path | None = None
    ):
        """
        Establish connection to RabbitMQ using pika.
        """
        try:
            from pika.adapters.select_connection import SelectConnection

            if config is None:
                if config_path is None and AuthConfig.default_config_exists():
                    auth_config = AuthConfig.from_local_config()
                elif Path(config_path).exists():
                    auth_config = AuthConfig.from_local_config(config_path=config_path)
                else:
                    logger.error("Fail to initialize auth_config")
                    return
                self.rabbitmq_config = auth_config.rabbitmq
            elif isinstance(config, RabbitMQConfig):
                self.rabbitmq_config = config
            elif isinstance(config, dict):
                self.rabbitmq_config = AuthConfig.from_dict(config).rabbitmq
            else:
                logger.error("Not implemented")

                # Start connection process
            parameters = self.get_rabbitmq_connection_param()
            self.rabbitmq_connection = SelectConnection(
                parameters,
                on_open_callback=self.on_rabbitmq_connection_open,
                on_open_error_callback=self.on_rabbitmq_connection_error,
                on_close_callback=self.on_rabbitmq_connection_closed,
            )

            # Start IOLoop in dedicated thread
            self._io_loop_thread = ContextThread(
                target=self.rabbitmq_connection.ioloop.start, daemon=True
            )
            self._io_loop_thread.start()
            logger.info("RabbitMQ connection process started")
        except Exception:
            logger.error("Fail to initialize auth_config", exc_info=True)

    def get_rabbitmq_queue_size(self) -> int:
        """Get the current number of messages in the queue.

        Returns:
            int: Number of messages in the queue.
                 Returns -1 if there's an error or no active connection.
        """
        if self.rabbitmq_exchange_type != DIRECT_EXCHANGE_TYPE:
            logger.warning("Queue size can only be checked for direct exchanges")
            return None

        with self._rabbitmq_lock:
            if not self.is_rabbitmq_connected():
                logger.warning("No active connection to check queue size")
                return -1

            # Declare queue passively (only checks existence, doesn't create)
            # Using passive=True prevents accidental queue creation
            result = self.rabbitmq_channel.queue_declare(
                queue=self.rabbit_queue_name,
                durable=True,  # Match the original queue durability setting
                passive=True,  # Only check queue existence, don't create
            )

            if result is None:
                return 0
            # Return the message count from the queue declaration result
            return result.method.message_count

    def get_rabbitmq_connection_param(self):
        import pika

        credentials = pika.PlainCredentials(
            username=self.rabbitmq_config.user_name,
            password=self.rabbitmq_config.password,
            erase_on_connect=self.rabbitmq_config.erase_on_connect,
        )
        if self.rabbitmq_config.port == 5671:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            context.check_hostname = False
            context.verify_mode = False
            return pika.ConnectionParameters(
                host=self.rabbitmq_config.host_name,
                port=self.rabbitmq_config.port,
                virtual_host=self.rabbitmq_config.virtual_host,
                credentials=credentials,
                ssl_options=pika.SSLOptions(context),
                connection_attempts=self.rabbitmq_connection_attempts,
                retry_delay=self.rabbitmq_retry_delay,
                heartbeat=self.rabbitmq_heartbeat,
            )
        else:
            return pika.ConnectionParameters(
                host=self.rabbitmq_config.host_name,
                port=self.rabbitmq_config.port,
                virtual_host=self.rabbitmq_config.virtual_host,
                credentials=credentials,
                connection_attempts=self.rabbitmq_connection_attempts,
                retry_delay=self.rabbitmq_retry_delay,
                heartbeat=self.rabbitmq_heartbeat,
            )

    # Connection lifecycle callbacks
    def on_rabbitmq_connection_open(self, connection):
        """Called when connection is established."""
        logger.debug("Connection opened")
        connection.channel(on_open_callback=self.on_rabbitmq_channel_open)

    def on_rabbitmq_connection_error(self, connection, error):
        """Called if connection fails to open."""
        logger.error(f"Connection failed: {error}")
        self.rabbit_reconnect()

    def on_rabbitmq_connection_closed(self, connection, reason):
        """Called when connection closes."""
        logger.warning(f"Connection closed: {reason}")
        if not self._rabbitmq_stop_flag:
            self.rabbit_reconnect()

    # Channel lifecycle callbacks
    def on_rabbitmq_channel_open(self, channel):
        """Called when channel is ready."""
        self.rabbitmq_channel = channel
        logger.debug("Channel opened")

        # Setup exchange and queue
        channel.exchange_declare(
            exchange=self.rabbitmq_exchange_name,
            exchange_type=self.rabbitmq_exchange_type,
            durable=True,
            callback=self.on_rabbitmq_exchange_declared,
        )

    def on_rabbitmq_exchange_declared(self, frame):
        """Called when exchange is ready."""
        self.rabbitmq_channel.queue_declare(
            queue=self.rabbit_queue_name, durable=True, callback=self.on_rabbitmq_queue_declared
        )

    def on_rabbitmq_queue_declared(self, frame):
        """Called when queue is ready."""
        self.rabbitmq_channel.queue_bind(
            exchange=self.rabbitmq_exchange_name,
            queue=self.rabbit_queue_name,
            routing_key=self.rabbit_queue_name,
            callback=self.on_rabbitmq_bind_ok,
        )

    def on_rabbitmq_bind_ok(self, frame):
        """Final setup step when bind is complete."""
        logger.info("RabbitMQ setup completed")

    def on_rabbitmq_message(self, channel, method, properties, body):
        """Handle incoming messages. Only for test."""
        try:
            print(f"Received message: {body.decode()}\n")
            self.rabbitmq_message_cache.put({"properties": properties, "body": body})
            print(f"message delivery_tag: {method.delivery_tag}\n")
            channel.basic_ack(delivery_tag=method.delivery_tag)
        except Exception as e:
            logger.error(f"Message handling failed: {e}", exc_info=True)

    def wait_for_connection_ready(self):
        start_time = time.time()
        while not self.is_rabbitmq_connected():
            delta_time = time.time() - start_time
            if delta_time > self.rabbitmq_conn_max_waiting_seconds:
                logger.error("Failed to start consuming: Connection timeout")
                return False
            self.rabbit_reconnect()
            time.sleep(self.rabbitmq_conn_sleep_seconds)  # Reduced frequency of checks

    # Message handling
    def rabbitmq_start_consuming(self):
        """Start consuming messages asynchronously."""
        self.wait_for_connection_ready()

        self.rabbitmq_channel.basic_consume(
            queue=self.rabbit_queue_name,
            on_message_callback=self.on_rabbitmq_message,
            auto_ack=False,
        )
        logger.info("Started rabbitmq consuming messages")

    def rabbitmq_publish_message(self, message: dict):
        """
        Publish a message to RabbitMQ.
        """
        import pika

        with self._rabbitmq_lock:
            if not self.is_rabbitmq_connected():
                logger.error("Cannot publish - no active connection")
                return False

            try:
                self.rabbitmq_channel.basic_publish(
                    exchange=self.rabbitmq_exchange_name,
                    routing_key=self.rabbit_queue_name,
                    body=json.dumps(message),
                    properties=pika.BasicProperties(
                        delivery_mode=2,  # Persistent
                    ),
                    mandatory=True,
                )
                logger.debug(f"Published message: {message}")
                return True
            except Exception as e:
                logger.error(f"Failed to publish message: {e}")
                self.rabbit_reconnect()
                return False

    # Connection management
    def rabbit_reconnect(self):
        """Schedule reconnection attempt."""
        logger.info("Attempting to reconnect...")
        if self.rabbitmq_connection and not self.rabbitmq_connection.is_closed:
            self.rabbitmq_connection.ioloop.stop()

        # Reset connection state
        self.rabbitmq_channel = None
        self.initialize_rabbitmq()

    def rabbitmq_close(self):
        """Gracefully shutdown connection."""
        with self._rabbitmq_lock:
            self._rabbitmq_stop_flag = True

            # Close channel if open
            if self.rabbitmq_channel and self.rabbitmq_channel.is_open:
                try:
                    self.rabbitmq_channel.close()
                except Exception as e:
                    logger.warning(f"Error closing channel: {e}")

            # Close connection if open
            if self.rabbitmq_connection:
                if self.rabbitmq_connection.is_open:
                    try:
                        self.rabbitmq_connection.close()
                    except Exception as e:
                        logger.warning(f"Error closing connection: {e}")

                # Stop IOLoop if running
                try:
                    self.rabbitmq_connection.ioloop.stop()
                except Exception as e:
                    logger.warning(f"Error stopping IOLoop: {e}")

            # Wait for IOLoop thread to finish
            if self._io_loop_thread and self._io_loop_thread.is_alive():
                self._io_loop_thread.join(timeout=5)
                if self._io_loop_thread.is_alive():
                    logger.warning("IOLoop thread did not terminate cleanly")

        logger.info("RabbitMQ connection closed")



from typing import Any, ClassVar

from memos.configs.mem_scheduler import SchedulerConfigFactory
from memos.mem_scheduler.base_scheduler import BaseScheduler
from memos.mem_scheduler.general_scheduler import GeneralScheduler
from memos.mem_scheduler.optimized_scheduler import OptimizedScheduler


class SchedulerFactory(BaseScheduler):
    """Factory class for creating scheduler instances."""

    backend_to_class: ClassVar[dict[str, Any]] = {
        "general_scheduler": GeneralScheduler,
        "optimized_scheduler": OptimizedScheduler,
    }

    @classmethod
    def from_config(cls, config_factory: SchedulerConfigFactory) -> GeneralScheduler:
        backend = config_factory.backend
        if backend not in cls.backend_to_class:
            raise ValueError(f"Invalid backend: {backend}")
        mem_scheduler_class = cls.backend_to_class[backend]
        return mem_scheduler_class(config_factory.config)



from typing import Any, ClassVar

from pydantic import Field, field_validator, model_validator

from memos.configs.base import BaseConfig


class BaseAdderConfig(BaseConfig):
    """Base configuration class for Adder."""


class NaiveAdderConfig(BaseAdderConfig):
    """Configuration for Naive Adder."""

    # No additional config needed since components are passed from parent


class AdderConfigFactory(BaseConfig):
    """Factory class for creating Adder configurations."""

    backend: str = Field(..., description="Backend for Adder")
    config: dict[str, Any] = Field(..., description="Configuration for the Adder backend")

    backend_to_class: ClassVar[dict[str, Any]] = {
        "naive": NaiveAdderConfig,
    }

    @field_validator("backend")
    @classmethod
    def validate_backend(cls, backend: str) -> str:
        """Validate the backend field."""
        if backend not in cls.backend_to_class:
            raise ValueError(f"Invalid backend: {backend}")
        return backend

    @model_validator(mode="after")
    def create_config(self) -> "AdderConfigFactory":
        config_class = self.backend_to_class[self.backend]
        self.config = config_class(**self.config)
        return self


class BaseExtractorConfig(BaseConfig):
    """Base configuration class for Extractor."""


class NaiveExtractorConfig(BaseExtractorConfig):
    """Configuration for Naive Extractor."""


class ExtractorConfigFactory(BaseConfig):
    """Factory class for creating Extractor configurations."""

    backend: str = Field(..., description="Backend for Extractor")
    config: dict[str, Any] = Field(..., description="Configuration for the Extractor backend")

    backend_to_class: ClassVar[dict[str, Any]] = {
        "naive": NaiveExtractorConfig,
    }

    @field_validator("backend")
    @classmethod
    def validate_backend(cls, backend: str) -> str:
        """Validate the backend field."""
        if backend not in cls.backend_to_class:
            raise ValueError(f"Invalid backend: {backend}")
        return backend

    @model_validator(mode="after")
    def create_config(self) -> "ExtractorConfigFactory":
        config_class = self.backend_to_class[self.backend]
        self.config = config_class(**self.config)
        return self


class BaseRetrieverConfig(BaseConfig):
    """Base configuration class for Retrievers."""


class NaiveRetrieverConfig(BaseRetrieverConfig):
    """Configuration for Naive Retriever."""


class RetrieverConfigFactory(BaseConfig):
    """Factory class for creating Retriever configurations."""

    backend: str = Field(..., description="Backend for Retriever")
    config: dict[str, Any] = Field(..., description="Configuration for the Retriever backend")

    backend_to_class: ClassVar[dict[str, Any]] = {
        "naive": NaiveRetrieverConfig,
    }

    @field_validator("backend")
    @classmethod
    def validate_backend(cls, backend: str) -> str:
        """Validate the backend field."""
        if backend not in cls.backend_to_class:
            raise ValueError(f"Invalid backend: {backend}")
        return backend

    @model_validator(mode="after")
    def create_config(self) -> "RetrieverConfigFactory":
        config_class = self.backend_to_class[self.backend]
        self.config = config_class(**self.config)
        return self



"""
Singleton decorator module for caching factory instances to avoid excessive memory usage
from repeated initialization.
"""

import hashlib
import json

from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar
from weakref import WeakValueDictionary


T = TypeVar("T")


class FactorySingleton:
    """Factory singleton manager that caches instances based on configuration parameters"""

    def __init__(self):
        # Use weak reference dictionary for automatic cleanup when instances are no longer referenced
        self._instances: dict[str, WeakValueDictionary] = {}

    def _generate_cache_key(self, config: Any, *args, **kwargs) -> str:
        """Generate cache key based on configuration only (ignoring other parameters)"""

        # Handle configuration objects - only use the config parameter
        if hasattr(config, "model_dump"):  # Pydantic model
            config_data = config.model_dump()
        elif hasattr(config, "dict"):  # Legacy Pydantic model
            config_data = config.dict()
        elif isinstance(config, dict):
            config_data = config
        else:
            # For other types, try to convert to string
            config_data = str(config)

        # Filter out time-related fields that shouldn't affect caching
        filtered_config = self._filter_temporal_fields(config_data)

        # Generate hash key based only on config
        try:
            cache_str = json.dumps(filtered_config, sort_keys=True, ensure_ascii=False, default=str)
        except (TypeError, ValueError):
            # If JSON serialization fails, convert the entire config to string
            cache_str = str(filtered_config)

        return hashlib.md5(cache_str.encode("utf-8")).hexdigest()

    def _filter_temporal_fields(self, config_data: Any) -> Any:
        """Filter out temporal fields that shouldn't affect instance caching"""
        if isinstance(config_data, dict):
            filtered = {}
            for key, value in config_data.items():
                # Skip common temporal field names
                if key.lower() in {
                    "created_at",
                    "updated_at",
                    "timestamp",
                    "time",
                    "date",
                    "created_time",
                    "updated_time",
                    "last_modified",
                    "modified_at",
                    "start_time",
                    "end_time",
                    "execution_time",
                    "run_time",
                }:
                    continue
                # Recursively filter nested dictionaries
                filtered[key] = self._filter_temporal_fields(value)
            return filtered
        elif isinstance(config_data, list):
            # Recursively filter lists
            return [self._filter_temporal_fields(item) for item in config_data]
        else:
            # For primitive types, return as-is
            return config_data

    def get_or_create(self, factory_class: type, cache_key: str, creator_func: Callable) -> Any:
        """Get or create instance"""
        class_name = factory_class.__name__

        if class_name not in self._instances:
            self._instances[class_name] = WeakValueDictionary()

        class_cache = self._instances[class_name]

        if cache_key in class_cache:
            return class_cache[cache_key]

        # Create new instance
        instance = creator_func()
        class_cache[cache_key] = instance
        return instance

    def clear_cache(self, factory_class: type | None = None):
        """Clear cache"""
        if factory_class:
            class_name = factory_class.__name__
            if class_name in self._instances:
                self._instances[class_name].clear()
        else:
            for cache in self._instances.values():
                cache.clear()


# Global singleton manager
_factory_singleton = FactorySingleton()


def singleton_factory(factory_class: type | str | None = None):
    """
    Factory singleton decorator

    Usage:
    @singleton_factory()
    def from_config(cls, config):
        return SomeClass(config)

    Or specify factory class:
    @singleton_factory(EmbedderFactory)
    def from_config(cls, config):
        return SomeClass(config)
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> T:
            # Determine factory class and config parameter
            target_factory_class = factory_class
            config = None

            # Simple logic: check if first parameter is a class or config
            if args:
                if hasattr(args[0], "__name__") and hasattr(args[0], "__module__"):
                    # First parameter is a class (cls), so this is a @classmethod
                    if target_factory_class is None:
                        target_factory_class = args[0]
                    config = args[1] if len(args) > 1 else None
                else:
                    # First parameter is config, so this is a @staticmethod
                    if target_factory_class is None:
                        raise ValueError(
                            "Factory class must be explicitly specified for static methods"
                        )
                    if isinstance(target_factory_class, str):
                        # Convert string to a mock class for caching purposes
                        class MockFactoryClass:
                            __name__ = target_factory_class

                        target_factory_class = MockFactoryClass
                    config = args[0]

            if config is None:
                # If no configuration parameter, call original function directly
                return func(*args, **kwargs)

            # Generate cache key based only on config
            cache_key = _factory_singleton._generate_cache_key(config)

            # Function to create instance
            def creator():
                return func(*args, **kwargs)

            # Get or create instance
            return _factory_singleton.get_or_create(target_factory_class, cache_key, creator)

        return wrapper

    return decorator



from typing import Any, ClassVar

from memos.configs.llm import LLMConfigFactory
from memos.llms.base import BaseLLM
from memos.llms.deepseek import DeepSeekLLM
from memos.llms.hf import HFLLM
from memos.llms.hf_singleton import HFSingletonLLM
from memos.llms.ollama import OllamaLLM
from memos.llms.openai import AzureLLM, OpenAILLM
from memos.llms.qwen import QwenLLM
from memos.llms.vllm import VLLMLLM
from memos.memos_tools.singleton import singleton_factory


class LLMFactory(BaseLLM):
    """Factory class for creating LLM instances."""

    backend_to_class: ClassVar[dict[str, Any]] = {
        "openai": OpenAILLM,
        "azure": AzureLLM,
        "ollama": OllamaLLM,
        "huggingface": HFLLM,
        "huggingface_singleton": HFSingletonLLM,  # Add singleton version
        "vllm": VLLMLLM,
        "qwen": QwenLLM,
        "deepseek": DeepSeekLLM,
    }

    @classmethod
    @singleton_factory()
    def from_config(cls, config_factory: LLMConfigFactory) -> BaseLLM:
        backend = config_factory.backend
        if backend not in cls.backend_to_class:
            raise ValueError(f"Invalid backend: {backend}")
        llm_class = cls.backend_to_class[backend]
        return llm_class(config_factory.config)



from typing import Any, cast

from memos.configs.llm import VLLMLLMConfig
from memos.llms.base import BaseLLM
from memos.llms.utils import remove_thinking_tags
from memos.log import get_logger
from memos.types import MessageDict


logger = get_logger(__name__)


class VLLMLLM(BaseLLM):
    """
    VLLM LLM class for connecting to existing vLLM servers.
    """

    def __init__(self, config: VLLMLLMConfig):
        """
        Initialize the VLLM LLM to connect to an existing vLLM server.
        """
        self.config = config

        # Initialize OpenAI client for API calls
        self.client = None
        api_key = getattr(self.config, "api_key", "dummy")
        if not api_key:
            api_key = "dummy"

        import openai

        self.client = openai.Client(
            api_key=api_key, base_url=getattr(self.config, "api_base", "http://localhost:8088/v1")
        )

    def build_vllm_kv_cache(self, messages: Any) -> str:
        """
        Build a KV cache from chat messages via one vLLM request.
        Handles str, list[str], and MessageList formats.
        """
        # 1. Normalize input to a MessageList
        processed_messages: list[MessageDict] = []
        if isinstance(messages, str):
            processed_messages = [
                {
                    "role": "system",
                    "content": f"Below is some information about the user.\n{messages}",
                }
            ]
        elif isinstance(messages, list):
            if not messages:
                pass  # Empty list
            elif isinstance(messages[0], str):
                str_content = " ".join(str(msg) for msg in messages)
                processed_messages = [
                    {
                        "role": "system",
                        "content": f"Below is some information about the user.\n{str_content}",
                    }
                ]
            elif isinstance(messages[0], dict):
                processed_messages = cast("list[MessageDict]", messages)

        # 2. Convert to prompt for logging/return value.
        prompt = self._messages_to_prompt(processed_messages)

        if not prompt.strip():
            raise ValueError("Prompt is empty, cannot build KV cache.")

        # 3. Send request to vLLM server to preload the KV cache
        if self.client:
            try:
                # Use the processed messages for the API call
                prefill_kwargs = {
                    "model": self.config.model_name_or_path,
                    "messages": processed_messages,
                    "max_tokens": 2,
                    "temperature": 0.0,
                    "top_p": 1.0,
                }
                self.client.chat.completions.create(**prefill_kwargs)
                logger.info(f"vLLM KV cache prefill completed for prompt: '{prompt[:100]}...'")
            except Exception as e:
                logger.warning(f"Failed to prefill vLLM KV cache: {e}")

        return prompt

    def generate(self, messages: list[MessageDict]) -> str:
        """
        Generate a response from the model.
        """
        if self.client:
            return self._generate_with_api_client(messages)
        else:
            raise RuntimeError("API client is not available")

    def _generate_with_api_client(self, messages: list[MessageDict]) -> str:
        """
        Generate response using vLLM API client.
        """
        if self.client:
            completion_kwargs = {
                "model": self.config.model_name_or_path,
                "messages": messages,
                "temperature": float(getattr(self.config, "temperature", 0.8)),
                "max_tokens": int(getattr(self.config, "max_tokens", 1024)),
                "top_p": float(getattr(self.config, "top_p", 0.9)),
                "extra_body": {"chat_template_kwargs": {"enable_thinking": False}},
            }

            response = self.client.chat.completions.create(**completion_kwargs)
            response_text = response.choices[0].message.content or ""
            logger.info(f"VLLM API response: {response_text}")
            return (
                remove_thinking_tags(response_text)
                if getattr(self.config, "remove_think_prefix", False)
                else response_text
            )
        else:
            raise RuntimeError("API client is not available")

    def _messages_to_prompt(self, messages: list[MessageDict]) -> str:
        """
        Convert messages to prompt string.
        """
        prompt_parts = []
        for msg in messages:
            role = msg["role"]
            content = msg["content"]
            prompt_parts.append(f"{role.capitalize()}: {content}")
        return "\n".join(prompt_parts)

    def generate_stream(self, messages: list[MessageDict]):
        """
        Generate a response from the model using streaming.
        Yields content chunks as they are received.
        """
        if self.client:
            completion_kwargs = {
                "model": self.config.model_name_or_path,
                "messages": messages,
                "temperature": float(getattr(self.config, "temperature", 0.8)),
                "max_tokens": int(getattr(self.config, "max_tokens", 1024)),
                "top_p": float(getattr(self.config, "top_p", 0.9)),
                "stream": True,  # Enable streaming
                "extra_body": {"chat_template_kwargs": {"enable_thinking": False}},
            }

            stream = self.client.chat.completions.create(**completion_kwargs)
            for chunk in stream:
                content = chunk.choices[0].delta.content
                if content:
                    yield content
        else:
            raise RuntimeError("API client is not available")



import json
import os

from collections import OrderedDict
from typing import TYPE_CHECKING, Any

from memos.api.product_models import APISearchRequest
from memos.configs.mem_scheduler import GeneralSchedulerConfig
from memos.log import get_logger
from memos.mem_cube.general import GeneralMemCube
from memos.mem_cube.navie import NaiveMemCube
from memos.mem_scheduler.general_modules.api_misc import SchedulerAPIModule
from memos.mem_scheduler.general_scheduler import GeneralScheduler
from memos.mem_scheduler.schemas.general_schemas import (
    API_MIX_SEARCH_LABEL,
    MemCubeID,
    SearchMode,
    UserID,
)
from memos.mem_scheduler.schemas.message_schemas import ScheduleMessageItem
from memos.mem_scheduler.utils.api_utils import format_textual_memory_item
from memos.mem_scheduler.utils.db_utils import get_utc_now
from memos.mem_scheduler.utils.misc_utils import group_messages_by_user_and_mem_cube
from memos.memories.textual.tree import TextualMemoryItem, TreeTextMemory
from memos.types import UserContext


if TYPE_CHECKING:
    from memos.mem_scheduler.schemas.monitor_schemas import MemoryMonitorItem

logger = get_logger(__name__)


class OptimizedScheduler(GeneralScheduler):
    """Optimized scheduler with improved working memory management and support for api"""

    def __init__(self, config: GeneralSchedulerConfig):
        super().__init__(config)
        self.window_size = int(os.getenv("API_SEARCH_WINDOW_SIZE", 5))
        self.history_memory_turns = int(os.getenv("API_SEARCH_HISTORY_TURNS", 5))
        self.session_counter = OrderedDict()
        self.max_session_history = 5

        self.api_module = SchedulerAPIModule(
            window_size=self.window_size,
            history_memory_turns=self.history_memory_turns,
        )
        self.register_handlers(
            {
                API_MIX_SEARCH_LABEL: self._api_mix_search_message_consumer,
            }
        )
        self.searcher = None
        self.reranker = None
        self.text_mem = None

    def submit_memory_history_async_task(
        self,
        search_req: APISearchRequest,
        user_context: UserContext,
        memories_to_store: dict | None = None,
        session_id: str | None = None,
    ):
        # Create message for async fine search
        message_content = {
            "search_req": {
                "query": search_req.query,
                "user_id": search_req.user_id,
                "session_id": session_id,
                "top_k": search_req.top_k,
                "internet_search": search_req.internet_search,
                "moscube": search_req.moscube,
                "chat_history": search_req.chat_history,
            },
            "user_context": {"mem_cube_id": user_context.mem_cube_id},
            "memories_to_store": memories_to_store,
        }

        async_task_id = f"mix_search_{search_req.user_id}_{get_utc_now().timestamp()}"

        message = ScheduleMessageItem(
            item_id=async_task_id,
            user_id=search_req.user_id,
            mem_cube_id=user_context.mem_cube_id,
            label=API_MIX_SEARCH_LABEL,
            content=json.dumps(message_content),
            timestamp=get_utc_now(),
        )

        # Submit async task
        self.memos_message_queue.submit_messages([message])
        logger.info(f"Submitted async fine search task for user {search_req.user_id}")
        return async_task_id

    def search_memories(
        self,
        search_req: APISearchRequest,
        user_context: UserContext,
        mem_cube: NaiveMemCube,
        mode: SearchMode,
    ):
        """Fine search memories function copied from server_router to avoid circular import"""
        target_session_id = search_req.session_id
        if not target_session_id:
            target_session_id = "default_session"
        search_filter = {"session_id": search_req.session_id} if search_req.session_id else None

        # Create MemCube and perform search
        search_results = mem_cube.text_mem.search(
            query=search_req.query,
            user_name=user_context.mem_cube_id,
            top_k=search_req.top_k,
            mode=mode,
            manual_close_internet=not search_req.internet_search,
            moscube=search_req.moscube,
            search_filter=search_filter,
            info={
                "user_id": search_req.user_id,
                "session_id": target_session_id,
                "chat_history": search_req.chat_history,
            },
        )
        return search_results

    def mix_search_memories(
        self,
        search_req: APISearchRequest,
        user_context: UserContext,
    ) -> list[dict[str, Any]]:
        """
        Mix search memories: fast search + async fine search
        """
        logger.info(
            f"Mix searching memories for user {search_req.user_id} with query: {search_req.query}"
        )

        # Get mem_cube for fast search
        target_session_id = search_req.session_id
        if not target_session_id:
            target_session_id = "default_session"
        search_filter = {"session_id": search_req.session_id} if search_req.session_id else None

        # Rerank Memories - reranker expects TextualMemoryItem objects

        info = {
            "user_id": search_req.user_id,
            "session_id": target_session_id,
            "chat_history": search_req.chat_history,
        }

        fast_retrieved_memories = self.searcher.retrieve(
            query=search_req.query,
            user_name=user_context.mem_cube_id,
            top_k=search_req.top_k,
            mode=SearchMode.FAST,
            manual_close_internet=not search_req.internet_search,
            moscube=search_req.moscube,
            search_filter=search_filter,
            info=info,
        )

        # Try to get pre-computed memories if available
        history_memories = self.api_module.get_history_memories(
            user_id=search_req.user_id,
            mem_cube_id=user_context.mem_cube_id,
            turns=self.history_memory_turns,
        )
        logger.info(f"Found {len(history_memories)} history memories.")
        if not history_memories:
            memories = self.searcher.post_retrieve(
                retrieved_results=fast_retrieved_memories,
                top_k=search_req.top_k,
                user_name=user_context.mem_cube_id,
                info=info,
            )
        else:
            # if history memories can directly answer
            sorted_history_memories = self.reranker.rerank(
                query=search_req.query,  # Use search_req.query instead of undefined query
                graph_results=history_memories,  # Pass TextualMemoryItem objects directly
                top_k=search_req.top_k,  # Use search_req.top_k instead of undefined top_k
                search_filter=search_filter,
            )
            logger.info(f"Reranked {len(sorted_history_memories)} history memories.")
            processed_hist_mem = self.searcher.post_retrieve(
                retrieved_results=sorted_history_memories,
                top_k=search_req.top_k,
                user_name=user_context.mem_cube_id,
                info=info,
            )

            can_answer = self.retriever.evaluate_memory_answer_ability(
                query=search_req.query, memory_texts=[one.memory for one in processed_hist_mem]
            )

            if can_answer:
                logger.info("History memories can answer the query.")
                sorted_results = fast_retrieved_memories + sorted_history_memories
                combined_results = self.searcher.post_retrieve(
                    retrieved_results=sorted_results,
                    top_k=search_req.top_k,
                    user_name=user_context.mem_cube_id,
                    info=info,
                )
                memories = combined_results[: search_req.top_k]
            else:
                logger.info("History memories cannot answer the query, enhancing memories.")
                sorted_results = fast_retrieved_memories + sorted_history_memories
                combined_results = self.searcher.post_retrieve(
                    retrieved_results=sorted_results,
                    top_k=search_req.top_k,
                    user_name=user_context.mem_cube_id,
                    info=info,
                )
                enhanced_memories, _ = self.retriever.enhance_memories_with_query(
                    query_history=[search_req.query],
                    memories=combined_results,
                )

                if len(enhanced_memories) < search_req.top_k:
                    logger.info(
                        f"Enhanced memories ({len(enhanced_memories)}) are less than top_k ({search_req.top_k}). Recalling for more."
                    )
                    missing_info_hint, trigger = self.retriever.recall_for_missing_memories(
                        query=search_req.query,
                        memories=combined_results,
                    )
                    retrieval_size = search_req.top_k - len(enhanced_memories)
                    if trigger:
                        logger.info(f"Triggering additional search with hint: {missing_info_hint}")
                        additional_memories = self.searcher.search(
                            query=missing_info_hint,
                            user_name=user_context.mem_cube_id,
                            top_k=retrieval_size,
                            mode=SearchMode.FAST,
                            memory_type="All",
                            search_filter=search_filter,
                            info=info,
                        )
                    else:
                        logger.info("Not triggering additional search, using combined results.")
                        additional_memories = combined_results[:retrieval_size]
                    logger.info(
                        f"Added {len(additional_memories)} more memories. Total enhanced memories: {len(enhanced_memories)}"
                    )
                    enhanced_memories += additional_memories

                memories = enhanced_memories[: search_req.top_k]

        formatted_memories = [format_textual_memory_item(item) for item in memories]
        logger.info("Submitted memory history async task.")
        self.submit_memory_history_async_task(
            search_req=search_req,
            user_context=user_context,
            memories_to_store={
                "memories": [one.to_dict() for one in memories],
                "formatted_memories": formatted_memories,
            },
        )

        return formatted_memories

    def update_search_memories_to_redis(
        self,
        messages: list[ScheduleMessageItem],
    ):
        for msg in messages:
            content_dict = json.loads(msg.content)
            search_req = content_dict["search_req"]
            user_context = content_dict["user_context"]
            session_id = search_req.get("session_id")
            if session_id:
                if session_id not in self.session_counter:
                    self.session_counter[session_id] = 0
                else:
                    self.session_counter[session_id] += 1
                session_turn = self.session_counter[session_id]

                # Move the current session to the end to mark it as recently used
                self.session_counter.move_to_end(session_id)

                # If the counter exceeds the max size, remove the oldest item
                if len(self.session_counter) > self.max_session_history:
                    self.session_counter.popitem(last=False)
            else:
                session_turn = 0

            memories_to_store = content_dict["memories_to_store"]
            if memories_to_store is None:
                memories: list[TextualMemoryItem] = self.search_memories(
                    search_req=APISearchRequest(**content_dict["search_req"]),
                    user_context=UserContext(**content_dict["user_context"]),
                    mem_cube=self.current_mem_cube,
                    mode=SearchMode.FAST,
                )
                formatted_memories = [format_textual_memory_item(data) for data in memories]
            else:
                memories = [
                    TextualMemoryItem.from_dict(one) for one in memories_to_store["memories"]
                ]
                formatted_memories = memories_to_store["formatted_memories"]

            # Sync search data to Redis
            self.api_module.sync_search_data(
                item_id=msg.item_id,
                user_id=search_req["user_id"],
                mem_cube_id=user_context["mem_cube_id"],
                query=search_req["query"],
                memories=memories,
                formatted_memories=formatted_memories,
                session_id=session_id,
                conversation_turn=session_turn,
            )

    def _api_mix_search_message_consumer(self, messages: list[ScheduleMessageItem]) -> None:
        """
        Process and handle query trigger messages from the queue.

        Args:
            messages: List of query messages to process
        """
        logger.info(f"Messages {messages} assigned to {API_MIX_SEARCH_LABEL} handler.")

        # Process the query in a session turn
        grouped_messages = group_messages_by_user_and_mem_cube(messages)

        self.validate_schedule_messages(messages=messages, label=API_MIX_SEARCH_LABEL)

        for user_id in grouped_messages:
            for mem_cube_id in grouped_messages[user_id]:
                messages = grouped_messages[user_id][mem_cube_id]
                if len(messages) == 0:
                    return
                self.update_search_memories_to_redis(messages=messages)

    def replace_working_memory(
        self,
        user_id: UserID | str,
        mem_cube_id: MemCubeID | str,
        mem_cube: GeneralMemCube,
        original_memory: list[TextualMemoryItem],
        new_memory: list[TextualMemoryItem],
    ) -> None | list[TextualMemoryItem]:
        """Replace working memory with new memories after reranking."""
        text_mem_base = mem_cube.text_mem
        if isinstance(text_mem_base, TreeTextMemory):
            text_mem_base: TreeTextMemory = text_mem_base

            # process rerank memories with llm
            query_db_manager = self.monitor.query_monitors[user_id][mem_cube_id]
            # Sync with database to get latest query history
            query_db_manager.sync_with_orm()

            query_history = query_db_manager.obj.get_queries_with_timesort()
            memories_with_new_order, rerank_success_flag = (
                self.retriever.process_and_rerank_memories(
                    queries=query_history,
                    original_memory=original_memory,
                    new_memory=new_memory,
                    top_k=self.top_k,
                )
            )

            # Apply combined filtering (unrelated + redundant)
            logger.info(
                f"Applying combined unrelated and redundant memory filtering to {len(memories_with_new_order)} memories"
            )
            filtered_memories, filtering_success_flag = (
                self.retriever.filter_unrelated_and_redundant_memories(
                    query_history=query_history,
                    memories=memories_with_new_order,
                )
            )

            if filtering_success_flag:
                logger.info(
                    f"Combined filtering completed successfully. "
                    f"Filtered from {len(memories_with_new_order)} to {len(filtered_memories)} memories"
                )
                memories_with_new_order = filtered_memories
            else:
                logger.warning(
                    "Combined filtering failed - keeping memories as fallback. "
                    f"Count: {len(memories_with_new_order)}"
                )

            # Update working memory monitors
            query_keywords = query_db_manager.obj.get_keywords_collections()
            logger.info(
                f"Processing {len(memories_with_new_order)} memories with {len(query_keywords)} query keywords"
            )
            new_working_memory_monitors = self.transform_working_memories_to_monitors(
                query_keywords=query_keywords,
                memories=memories_with_new_order,
            )

            if not rerank_success_flag:
                for one in new_working_memory_monitors:
                    one.sorting_score = 0

            logger.info(f"update {len(new_working_memory_monitors)} working_memory_monitors")
            self.monitor.update_working_memory_monitors(
                new_working_memory_monitors=new_working_memory_monitors,
                user_id=user_id,
                mem_cube_id=mem_cube_id,
                mem_cube=mem_cube,
            )

            # Use the filtered and reranked memories directly
            text_mem_base.replace_working_memory(memories=memories_with_new_order)

            # Update monitor after replacing working memory
            mem_monitors: list[MemoryMonitorItem] = self.monitor.working_memory_monitors[user_id][
                mem_cube_id
            ].obj.get_sorted_mem_monitors(reverse=True)
            new_working_memories = [mem_monitor.tree_memory_item for mem_monitor in mem_monitors]

            logger.info(
                f"The working memory has been replaced with {len(memories_with_new_order)} new memories."
            )
            self.log_working_memory_replacement(
                original_memory=original_memory,
                new_memory=new_working_memories,
                user_id=user_id,
                mem_cube_id=mem_cube_id,
                mem_cube=mem_cube,
                log_func_callback=self._submit_web_logs,
            )
        else:
            logger.error("memory_base is not supported")
            memories_with_new_order = new_memory

        return memories_with_new_order
